import os
import sys
sys.path.insert(0, os.path.abspath('.'))
from idtap.classes.pitch import Pitch
import math
import pytest

# All tests mirror src/js/tests/pitch.test.ts

def test_default_pitch():
    p = Pitch()
    assert isinstance(p, Pitch)
    assert p.swara == 0
    assert p.oct == 0
    assert p.raised is True
    assert p.fundamental == 261.63
    ratios = [
        1,
        [2 ** (1 / 12), 2 ** (2 / 12)],
        [2 ** (3 / 12), 2 ** (4 / 12)],
        [2 ** (5 / 12), 2 ** (6 / 12)],
        2 ** (7 / 12),
        [2 ** (8 / 12), 2 ** (9 / 12)],
        [2 ** (10 / 12), 2 ** (11 / 12)]
    ]
    assert p.ratios == ratios
    assert p.log_offset == 0
    assert p.frequency == pytest.approx(261.63)
    assert p.non_offset_frequency == pytest.approx(261.63)
    log_freq = math.log2(261.63)
    assert p.non_offset_log_freq == pytest.approx(log_freq)
    assert p.log_freq == pytest.approx(log_freq)
    assert p.sargam_letter == 'S'
    assert p.octaved_sargam_letter == 'S'
    assert p.numbered_pitch == 0
    assert p.chroma == 0
    assert p.to_json() == {
        'swara': 0,
        'raised': True,
        'oct': 0,
        'ratios': ratios,
        'fundamental': 261.63,
        'logOffset': 0,
    }


def test_swara_input():
    def sa_test(p: Pitch):
        sa_freq = 261.63
        sa_log_freq = math.log2(sa_freq)
        assert p.swara == 0
        assert math.isclose(p.frequency, sa_freq, abs_tol=0.01)
        assert math.isclose(p.log_freq, sa_log_freq, abs_tol=0.01)
        assert p.sargam_letter == 'S'
        assert p.raised is True
        assert p.chroma == 0

    def re_lowered_test(p: Pitch):
        re_freq = 277.19
        re_log_freq = math.log2(re_freq)
        assert p.swara == 1
        assert math.isclose(p.frequency, re_freq, abs_tol=0.01)
        assert math.isclose(p.log_freq, re_log_freq, abs_tol=0.01)
        assert p.sargam_letter == 'r'
        assert p.raised is False
        assert p.chroma == 1

    def re_raised_test(p: Pitch):
        re_freq = 293.67
        re_log_freq = math.log2(re_freq)
        assert p.swara == 1
        assert math.isclose(p.frequency, re_freq, abs_tol=0.01)
        assert math.isclose(p.log_freq, re_log_freq, abs_tol=0.01)
        assert p.sargam_letter == 'R'
        assert p.raised is True
        assert p.chroma == 2

    def ga_lowered_test(p: Pitch):
        ga_freq = 311.13
        ga_log_freq = math.log2(ga_freq)
        assert p.swara == 2
        assert math.isclose(p.frequency, ga_freq, abs_tol=0.01)
        assert math.isclose(p.log_freq, ga_log_freq, abs_tol=0.01)
        assert p.sargam_letter == 'g'
        assert p.raised is False
        assert p.chroma == 3

    def ga_raised_test(p: Pitch):
        ga_freq = 329.63
        ga_log_freq = math.log2(ga_freq)
        assert p.swara == 2
        assert math.isclose(p.frequency, ga_freq, abs_tol=0.01)
        assert math.isclose(p.log_freq, ga_log_freq, abs_tol=0.01)
        assert p.sargam_letter == 'G'
        assert p.raised is True
        assert p.chroma == 4

    def ma_lowered_test(p: Pitch):
        ma_freq = 349.23
        ma_log_freq = math.log2(ma_freq)
        assert p.swara == 3
        assert math.isclose(p.frequency, ma_freq, abs_tol=0.01)
        assert math.isclose(p.log_freq, ma_log_freq, abs_tol=0.01)
        assert p.sargam_letter == 'm'
        assert p.raised is False
        assert p.chroma == 5

    def ma_raised_test(p: Pitch):
        ma_freq = 370
        ma_log_freq = math.log2(ma_freq)
        assert p.swara == 3
        assert math.isclose(p.frequency, ma_freq, abs_tol=0.01)
        assert math.isclose(p.log_freq, ma_log_freq, abs_tol=0.01)
        assert p.sargam_letter == 'M'
        assert p.raised is True
        assert p.chroma == 6

    def pa_test(p: Pitch):
        pa_freq = 392
        pa_log_freq = math.log2(pa_freq)
        assert p.swara == 4
        assert math.isclose(p.frequency, pa_freq, abs_tol=0.01)
        assert math.isclose(p.log_freq, pa_log_freq, abs_tol=0.01)
        assert p.sargam_letter == 'P'
        assert p.raised is True
        assert p.chroma == 7

    def dha_lowered_test(p: Pitch):
        dha_freq = 415.31
        dha_log_freq = math.log2(dha_freq)
        assert p.swara == 5
        assert math.isclose(p.frequency, dha_freq, abs_tol=0.01)
        assert math.isclose(p.log_freq, dha_log_freq, abs_tol=0.01)
        assert p.sargam_letter == 'd'
        assert p.raised is False
        assert p.chroma == 8

    def dha_raised_test(p: Pitch):
        dha_freq = 440.01
        dha_log_freq = math.log2(dha_freq)
        assert p.swara == 5
        assert math.isclose(p.frequency, dha_freq, abs_tol=0.01)
        assert math.isclose(p.log_freq, dha_log_freq, abs_tol=0.01)
        assert p.sargam_letter == 'D'
        assert p.raised is True
        assert p.chroma == 9

    def ni_lowered_test(p: Pitch):
        ni_freq = 466.17
        ni_log_freq = math.log2(ni_freq)
        assert p.swara == 6
        assert math.isclose(p.frequency, ni_freq, abs_tol=0.01)
        assert math.isclose(p.log_freq, ni_log_freq, abs_tol=0.01)
        assert p.sargam_letter == 'n'
        assert p.raised is False
        assert p.chroma == 10

    def ni_raised_test(p: Pitch):
        ni_freq = 493.89
        ni_log_freq = math.log2(ni_freq)
        assert p.swara == 6
        assert math.isclose(p.frequency, ni_freq, abs_tol=0.01)
        assert math.isclose(p.log_freq, ni_log_freq, abs_tol=0.01)
        assert p.sargam_letter == 'N'
        assert p.raised is True
        assert p.chroma == 11

    sa_vars = ['Sa', 'sa', 'S', 's', 0]
    for sa in sa_vars:
        p = Pitch({'swara': sa})
        sa_test(p)
        p = Pitch({'swara': sa, 'raised': False})
        sa_test(p)

    re_vars = ['Re', 're', 'R', 'r', 1]
    for re in re_vars:
        p = Pitch({'swara': re})
        re_raised_test(p)
        p = Pitch({'swara': re, 'raised': False})
        re_lowered_test(p)

    ga_vars = ['Ga', 'ga', 'G', 'g', 2]
    for ga in ga_vars:
        p = Pitch({'swara': ga})
        ga_raised_test(p)
        p = Pitch({'swara': ga, 'raised': False})
        ga_lowered_test(p)

    ma_vars = ['Ma', 'ma', 'M', 'm', 3]
    for ma in ma_vars:
        p = Pitch({'swara': ma})
        ma_raised_test(p)
        p = Pitch({'swara': ma, 'raised': False})
        ma_lowered_test(p)

    pa_vars = ['Pa', 'pa', 'P', 'p', 4]
    for pa in pa_vars:
        p = Pitch({'swara': pa})
        pa_test(p)
        p = Pitch({'swara': pa, 'raised': False})
        pa_test(p)

    dha_vars = ['Dha', 'dha', 'D', 'd', 5]
    for dha in dha_vars:
        p = Pitch({'swara': dha})
        dha_raised_test(p)
        p = Pitch({'swara': dha, 'raised': False})
        dha_lowered_test(p)

    ni_vars = ['Ni', 'ni', 'N', 'n', 6]
    for ni in ni_vars:
        p = Pitch({'swara': ni})
        ni_raised_test(p)
        p = Pitch({'swara': ni, 'raised': False})
        ni_lowered_test(p)


def test_octave_input():
    p = Pitch({'swara': 'sa', 'oct': -2})
    sa_down2 = 'S' + '\u0324'
    sa_down1 = 'S' + '\u0323'
    sa_plus1 = 'S' + '\u0307'
    sa_plus2 = 'S' + '\u0308'
    assert p.oct == -2
    assert p.octaved_sargam_letter == sa_down2
    p.set_oct(-1)
    assert p.oct == -1
    assert p.octaved_sargam_letter == sa_down1
    p.set_oct(0)
    assert p.oct == 0
    assert p.octaved_sargam_letter == 'S'
    p.set_oct(1)
    assert p.oct == 1
    assert p.octaved_sargam_letter == sa_plus1
    p.set_oct(2)
    assert p.oct == 2
    assert p.octaved_sargam_letter == sa_plus2


def test_log_offset():
    offset = 0.1
    p = Pitch({'log_offset': offset})
    assert p.log_offset == offset
    sa_freq = 261.63
    sa_log_freq = math.log2(sa_freq)
    offset_sa_log_freq = sa_log_freq + offset
    offset_sa_freq = 2 ** offset_sa_log_freq
    assert math.isclose(p.frequency, offset_sa_freq, abs_tol=0.01)
    assert math.isclose(p.log_freq, offset_sa_log_freq, abs_tol=0.01)
    assert math.isclose(p.non_offset_frequency, sa_freq, abs_tol=0.01)


def test_numbered_pitch():
    p = Pitch({'swara': 5, 'oct': -2})
    assert p.numbered_pitch == -15
    p = Pitch({'swara': 2, 'oct': 0})
    assert p.numbered_pitch == 4
    p = Pitch({'swara': 3, 'raised': False, 'oct': 1})
    assert p.numbered_pitch == 17


def test_same_as():
    p1 = Pitch({'swara': 're', 'raised': False, 'oct': 1})
    p2 = Pitch({'swara': 1, 'raised': False, 'oct': 1})
    p3 = Pitch({'swara': 1, 'raised': True, 'oct': 1})
    assert p1.same_as(p2)
    assert not p1.same_as(p3)


def test_from_pitch_number_and_helpers():
    p = Pitch.from_pitch_number(4)
    assert p.swara == 2
    assert p.raised is True
    assert p.oct == 0

    p = Pitch.from_pitch_number(-1)
    assert p.swara == 6
    assert p.raised is True
    assert p.oct == -1

    assert Pitch.pitch_number_to_chroma(14) == 2
    assert Pitch.pitch_number_to_chroma(-1) == 11

    sd, raised = Pitch.chroma_to_scale_degree(3)
    assert sd == 2
    assert raised is False
    sd, raised = Pitch.chroma_to_scale_degree(11)
    assert sd == 6
    assert raised is True


def test_display_properties():
    p_down = Pitch({'swara': 'g', 'raised': False, 'oct': -1})
    assert p_down.solfege_letter == 'Me'
    assert p_down.octaved_scale_degree == '3\u0323'
    assert p_down.octaved_solfege_letter == 'Me\u0323'
    assert p_down.octaved_solfege_letter_with_cents == 'Me\u0323 (+0\u00A2)'
    assert p_down.octaved_chroma == '3\u0323'
    assert p_down.octaved_chroma_with_cents == '3\u0323 (+0\u00A2)'
    assert p_down.cents_string == '+0\u00A2'
    assert p_down.a440_cents_deviation == 'D#3 (+0\u00A2)'
    assert p_down.movable_c_cents_deviation == 'D# (+0\u00A2)'

    p_up = Pitch({'swara': 'Sa', 'oct': 2})
    assert p_up.solfege_letter == 'Do'
    assert p_up.octaved_scale_degree == '1\u0308'
    assert p_up.octaved_solfege_letter == 'Do\u0308'
    assert p_up.octaved_solfege_letter_with_cents == 'Do\u0308 (+0\u00A2)'
    assert p_up.octaved_chroma == '0\u0308'
    assert p_up.octaved_chroma_with_cents == '0\u0308 (+0\u00A2)'
    assert p_up.cents_string == '+0\u00A2'
    assert p_up.a440_cents_deviation == 'C6 (+0\u00A2)'
    assert p_up.movable_c_cents_deviation == 'C (+0\u00A2)'


def test_frequency_and_set_oct_error_handling():
    p1 = Pitch()
    p1.swara = 0
    p1.ratios[0] = 'bad'
    with pytest.raises(SyntaxError):
        _ = p1.frequency
    with pytest.raises(SyntaxError):
        p1.set_oct(1)

    p2 = Pitch()
    p2.swara = 're'  # type: ignore
    with pytest.raises(SyntaxError):
        _ = p2.frequency
    with pytest.raises(SyntaxError):
        p2.set_oct(0)

    p3 = Pitch()
    p3.swara = 1
    p3.ratios[1] = 0
    with pytest.raises(SyntaxError):
        _ = p3.frequency


def test_formatted_string_getters_across_octaves():
    expected = {
        -2: 'C2 (+0\u00A2)',
        -1: 'C3 (+0\u00A2)',
        0: 'C4 (+0\u00A2)',
        1: 'C5 (+0\u00A2)',
        2: 'C6 (+0\u00A2)'
    }
    for i in range(-2, 3):
        p = Pitch({'swara': 'Sa', 'oct': i})
        assert p.a440_cents_deviation == expected[i]
        assert p.movable_c_cents_deviation == 'C (+0\u00A2)'


def test_chroma_to_scale_degree_all_mappings():
    expected = [
        (0, True),
        (1, False),
        (1, True),
        (2, False),
        (2, True),
        (3, False),
        (3, True),
        (4, True),
        (5, False),
        (5, True),
        (6, False),
        (6, True),
    ]
    for c in range(12):
        sd, raised = Pitch.chroma_to_scale_degree(c)
        assert sd == expected[c][0]
        assert raised == expected[c][1]


def test_numbered_pitch_edge_cases():
    low = Pitch({'swara': 'Sa', 'oct': -3})
    assert low.numbered_pitch == -36
    high = Pitch({'swara': 'ni', 'raised': True, 'oct': 3})
    assert high.numbered_pitch == 47
    bad = Pitch()
    bad.swara = 7
    with pytest.raises(SyntaxError):
        _ = bad.numbered_pitch


def test_constructor_error_conditions():
    with pytest.raises(SyntaxError):
        Pitch({'raised': 1})
    with pytest.raises(SyntaxError):
        Pitch({'swara': []})
    with pytest.raises(SyntaxError):
        Pitch({'swara': 'foo'})
    with pytest.raises(SyntaxError):
        Pitch({'oct': 0.5})
    with pytest.raises(SyntaxError):
        Pitch({'oct': '1'})
    with pytest.raises(SyntaxError):
        Pitch({'fundamental': 'A4'})
    with pytest.raises(SyntaxError):
        Pitch({'swara': 'x'})
    with pytest.raises(SyntaxError):
        Pitch({'swara': -1})
    with pytest.raises(SyntaxError):
        Pitch({'swara': 7})


def test_set_oct_invalid_inputs():
    bad_sa = Pitch()
    bad_sa.swara = 0
    bad_sa.ratios[0] = 'bad'
    with pytest.raises(SyntaxError):
        bad_sa.set_oct(1)

    bad_pa = Pitch({'swara': 'pa'})
    bad_pa.ratios[4] = None  # type: ignore
    with pytest.raises(SyntaxError):
        bad_pa.set_oct(0)

    bad_nested = Pitch({'swara': 're'})
    bad_nested.swara = 1
    bad_nested.ratios[1] = 0
    with pytest.raises(SyntaxError):
        bad_nested.set_oct(2)

    wrong_swara_type = Pitch()
    wrong_swara_type.swara = 're'  # type: ignore
    with pytest.raises(SyntaxError):
        wrong_swara_type.set_oct(0)


def test_non_offset_frequency_and_formatted_getters():
    sa = Pitch({'swara': 'Sa', 'log_offset': 0.1})
    assert sa.non_offset_frequency == pytest.approx(261.63)
    assert sa.non_offset_log_freq == pytest.approx(math.log2(261.63))
    assert sa.cents_string == '+120\u00A2'
    assert sa.a440_cents_deviation == 'C#4 (+20\u00A2)'
    assert sa.movable_c_cents_deviation == 'C (+120\u00A2)'
    assert sa.octaved_sargam_letter_with_cents == 'S (+120\u00A2)'

    ga = Pitch({'swara': 'ga', 'raised': False, 'log_offset': -0.05})
    ga_base = 261.63 * (2 ** (3/12))
    assert ga.non_offset_frequency == pytest.approx(ga_base)
    assert ga.non_offset_log_freq == pytest.approx(math.log2(ga_base))
    assert ga.cents_string == '-60\u00A2'
    assert ga.a440_cents_deviation == 'D4 (+40\u00A2)'
    assert ga.movable_c_cents_deviation == 'D# (-60\u00A2)'
    assert ga.octaved_sargam_letter_with_cents == 'g (-60\u00A2)'


def test_serialization_round_trip():
    p = Pitch({'swara': 'ga', 'raised': False, 'oct': 1, 'log_offset': 0.2})
    json_obj = p.to_json()
    copy = Pitch.from_json(json_obj)
    assert copy.to_json() == json_obj


def test_a440_cents_deviation_edge_octaves():
    expected = {
        -3: 'C1 (+0\u00A2)',
        -2: 'C2 (+0\u00A2)',
        -1: 'C3 (+0\u00A2)',
        0: 'C4 (+0\u00A2)',
        1: 'C5 (+0\u00A2)',
        2: 'C6 (+0\u00A2)',
        3: 'C7 (+0\u00A2)'
    }
    for i in range(-3, 4):
        p = Pitch({'swara': 'Sa', 'oct': i})
        assert p.a440_cents_deviation == expected[i]
        assert p.movable_c_cents_deviation == 'C (+0\u00A2)'


def test_octaved_display_strings_extreme_octaves():
    low = Pitch({'swara': 'Sa', 'oct': -3})
    high = Pitch({'swara': 'Sa', 'oct': 3})
    assert low.octaved_sargam_letter == 'S\u20E8'
    assert high.octaved_sargam_letter == 'S\u20DB'
    assert low.octaved_solfege_letter == 'Do\u20E8'
    assert high.octaved_solfege_letter == 'Do\u20DB'
    assert low.octaved_chroma == '0\u20E8'
    assert high.octaved_chroma == '0\u20DB'


def test_numbered_pitch_invalid_swara_values():
    p = Pitch()
    p.swara = -1
    with pytest.raises(SyntaxError):
        _ = p.numbered_pitch
    p.swara = 7
    with pytest.raises(SyntaxError):
        _ = p.numbered_pitch
    p.swara = 'ni'  # type: ignore
    with pytest.raises(SyntaxError):
        _ = p.numbered_pitch


def test_to_json_from_json_preserves_log_offset():
    orig = Pitch({'swara': 'ni', 'raised': False, 'oct': 2, 'log_offset': -0.3})
    round_trip = Pitch.from_json(orig.to_json())
    assert round_trip.to_json() == orig.to_json()
    assert math.isclose(round_trip.frequency, orig.frequency, abs_tol=0.0001)


def test_invalid_ratio_values_trigger_errors():
    bad_re = Pitch({'swara': 're'})
    bad_re.swara = 1
    bad_re.ratios[1] = 'bad'
    with pytest.raises(SyntaxError):
        _ = bad_re.frequency
    with pytest.raises(SyntaxError):
        bad_re.set_oct(1)

    bad_ga = Pitch({'swara': 'ga'})
    bad_ga.swara = 2
    bad_ga.ratios[2] = 5
    with pytest.raises(SyntaxError):
        _ = bad_ga.frequency
    with pytest.raises(SyntaxError):
        bad_ga.set_oct(0)


def test_western_pitch():
    p = Pitch({'swara': 're', 'raised': True})
    assert p.western_pitch == 'D'


def test_a440_cents_deviation_over_50():
    re = Pitch({'swara': 're', 'raised': True, 'log_offset': 0.05})
    assert re.a440_cents_deviation == 'E4 (-40\u00A2)'

    ni = Pitch({'swara': 'ni', 'raised': True, 'log_offset': 0.05})
    assert ni.a440_cents_deviation == 'C#4 (-40\u00A2)'


def test_constructor_rejects_undefined_ratios():
    base_ratios = [
        1,
        [2 ** (1 / 12), 2 ** (2 / 12)],
        [2 ** (3 / 12), 2 ** (4 / 12)],
        [2 ** (5 / 12), 2 ** (6 / 12)],
        2 ** (7 / 12),
        [2 ** (8 / 12), 2 ** (9 / 12)],
        [2 ** (10 / 12), 2 ** (11 / 12)]
    ]

    ratios1 = base_ratios.copy()
    ratios1[0] = None
    with pytest.raises(SyntaxError):
        Pitch({'ratios': ratios1})

    ratios2 = base_ratios.copy()
    ratios2[1] = [2 ** (1 / 12), None]
    with pytest.raises(SyntaxError):
        Pitch({'ratios': ratios2})


def test_latex_sargam_letter_basic():
    """Test that latex_sargam_letter returns the same as sargam_letter."""
    # Test all sargam letters in both raised and lowered forms
    sargam_tests = [
        ({'swara': 'sa'}, 'S'),
        ({'swara': 're', 'raised': False}, 'r'),
        ({'swara': 're', 'raised': True}, 'R'),
        ({'swara': 'ga', 'raised': False}, 'g'),
        ({'swara': 'ga', 'raised': True}, 'G'),
        ({'swara': 'ma', 'raised': False}, 'm'),
        ({'swara': 'ma', 'raised': True}, 'M'),
        ({'swara': 'pa'}, 'P'),
        ({'swara': 'dha', 'raised': False}, 'd'),
        ({'swara': 'dha', 'raised': True}, 'D'),
        ({'swara': 'ni', 'raised': False}, 'n'),
        ({'swara': 'ni', 'raised': True}, 'N'),
    ]
    
    for options, expected in sargam_tests:
        p = Pitch(options)
        assert p.latex_sargam_letter == expected
        assert p.latex_sargam_letter == p.sargam_letter


def test_latex_octaved_sargam_letter_no_octave():
    """Test LaTeX octaved sargam letter with no octave marking (oct=0)."""
    p = Pitch({'swara': 'sa', 'oct': 0})
    assert p.latex_octaved_sargam_letter == 'S'
    
    p = Pitch({'swara': 're', 'raised': False, 'oct': 0})
    assert p.latex_octaved_sargam_letter == 'r'


def test_latex_octaved_sargam_letter_positive_octaves():
    """Test LaTeX octaved sargam letter with positive octaves (dots above)."""
    # Test oct=1 (single dot above)
    p = Pitch({'swara': 'sa', 'oct': 1})
    assert p.latex_octaved_sargam_letter == r'$\dot{\mathrm{S}}$'
    
    p = Pitch({'swara': 're', 'raised': False, 'oct': 1})
    assert p.latex_octaved_sargam_letter == r'$\dot{\mathrm{r}}$'
    
    p = Pitch({'swara': 'ga', 'raised': True, 'oct': 1})
    assert p.latex_octaved_sargam_letter == r'$\dot{\mathrm{G}}$'
    
    # Test oct=2 (double dot above)
    p = Pitch({'swara': 'ma', 'raised': False, 'oct': 2})
    assert p.latex_octaved_sargam_letter == r'$\ddot{\mathrm{m}}$'
    
    p = Pitch({'swara': 'pa', 'oct': 2})
    assert p.latex_octaved_sargam_letter == r'$\ddot{\mathrm{P}}$'
    
    # Test oct=3 (triple dot above)
    p = Pitch({'swara': 'dha', 'raised': True, 'oct': 3})
    assert p.latex_octaved_sargam_letter == r'$\dddot{\mathrm{D}}$'
    
    p = Pitch({'swara': 'ni', 'raised': False, 'oct': 3})
    assert p.latex_octaved_sargam_letter == r'$\dddot{\mathrm{n}}$'


def test_latex_octaved_sargam_letter_negative_octaves():
    """Test LaTeX octaved sargam letter with negative octaves (dots below)."""
    # Test oct=-1 (single dot below)
    p = Pitch({'swara': 'sa', 'oct': -1})
    assert p.latex_octaved_sargam_letter == r'$\underset{\cdot}{\mathrm{S}}$'
    
    p = Pitch({'swara': 're', 'raised': True, 'oct': -1})
    assert p.latex_octaved_sargam_letter == r'$\underset{\cdot}{\mathrm{R}}$'
    
    # Test oct=-2 (double dot below)
    p = Pitch({'swara': 'ga', 'raised': False, 'oct': -2})
    assert p.latex_octaved_sargam_letter == r'$\underset{\cdot\cdot}{\mathrm{g}}$'
    
    p = Pitch({'swara': 'ma', 'raised': True, 'oct': -2})
    assert p.latex_octaved_sargam_letter == r'$\underset{\cdot\cdot}{\mathrm{M}}$'
    
    # Test oct=-3 (triple dot below)
    p = Pitch({'swara': 'pa', 'oct': -3})
    assert p.latex_octaved_sargam_letter == r'$\underset{\cdot\cdot\cdot}{\mathrm{P}}$'
    
    p = Pitch({'swara': 'dha', 'raised': False, 'oct': -3})
    assert p.latex_octaved_sargam_letter == r'$\underset{\cdot\cdot\cdot}{\mathrm{d}}$'


def test_latex_octaved_sargam_letter_all_sargam_all_octaves():
    """Test all sargam letters across all octave levels."""
    sargam_letters = ['sa', 're', 'ga', 'ma', 'pa', 'dha', 'ni']
    octave_expected = {
        -3: r'\underset{\cdot\cdot\cdot}',
        -2: r'\underset{\cdot\cdot}',
        -1: r'\underset{\cdot}',
        0: '',
        1: r'\dot',
        2: r'\ddot',
        3: r'\dddot'
    }
    
    for swara in sargam_letters:
        for raised in [True, False]:
            # Skip invalid combinations (sa and pa are always raised)
            if swara in ['sa', 'pa'] and not raised:
                continue
                
            p = Pitch({'swara': swara, 'raised': raised})
            base_letter = p.sargam_letter
            
            for oct in range(-3, 4):
                p_oct = Pitch({'swara': swara, 'raised': raised, 'oct': oct})
                expected_latex = octave_expected[oct]
                
                if oct == 0:
                    expected_result = base_letter
                elif expected_latex.startswith(r'\underset'):
                    expected_result = f'${expected_latex}{{\\mathrm{{{base_letter}}}}}$'
                else:
                    expected_result = f'${expected_latex}{{\\mathrm{{{base_letter}}}}}$'
                
                assert p_oct.latex_octaved_sargam_letter == expected_result


def test_latex_properties_preserve_backward_compatibility():
    """Test that existing properties are not affected by LaTeX additions."""
    test_cases = [
        {'swara': 'sa', 'oct': 0},
        {'swara': 're', 'raised': False, 'oct': 1},
        {'swara': 'ga', 'raised': True, 'oct': -1},
        {'swara': 'ma', 'raised': False, 'oct': 2},
        {'swara': 'pa', 'oct': -2},
        {'swara': 'dha', 'raised': True, 'oct': 3},
        {'swara': 'ni', 'raised': False, 'oct': -3},
    ]
    
    for options in test_cases:
        p = Pitch(options)
        
        # All existing properties should work exactly as before
        assert hasattr(p, 'sargam_letter')
        assert hasattr(p, 'octaved_sargam_letter')
        assert hasattr(p, 'frequency')
        assert hasattr(p, 'numbered_pitch')
        assert hasattr(p, 'chroma')
        
        # New LaTeX properties should be available
        assert hasattr(p, 'latex_sargam_letter')
        assert hasattr(p, 'latex_octaved_sargam_letter')
        
        # latex_sargam_letter should match sargam_letter
        assert p.latex_sargam_letter == p.sargam_letter


def test_latex_octave_diacritic_helper():
    """Test the _octave_latex_diacritic helper method."""
    # Test all octave levels
    octave_mapping = {
        -3: r'\underset{\cdot\cdot\cdot}',
        -2: r'\underset{\cdot\cdot}',
        -1: r'\underset{\cdot}',
        0: '',
        1: r'\dot',
        2: r'\ddot',
        3: r'\dddot'
    }
    
    for oct, expected in octave_mapping.items():
        p = Pitch({'swara': 'sa', 'oct': oct})
        assert p._octave_latex_diacritic() == expected


def test_latex_properties_edge_cases():
    """Test LaTeX properties with edge cases and various combinations."""
    # Test with log_offset (should not affect LaTeX output)
    p = Pitch({'swara': 'ga', 'raised': False, 'oct': 1, 'log_offset': 0.1})
    assert p.latex_octaved_sargam_letter == r'$\dot{\mathrm{g}}$'
    
    # Test with different fundamentals (should not affect LaTeX output)
    p = Pitch({'swara': 'ma', 'raised': True, 'oct': -1, 'fundamental': 440.0})
    assert p.latex_octaved_sargam_letter == r'$\underset{\cdot}{\mathrm{M}}$'
    
    # Test serialization includes existing functionality
    p = Pitch({'swara': 'dha', 'raised': False, 'oct': 2})
    json_data = p.to_json()
    p_restored = Pitch.from_json(json_data)
    
    # LaTeX properties should work after deserialization
    assert p_restored.latex_sargam_letter == 'd'
    assert p_restored.latex_octaved_sargam_letter == r'$\ddot{\mathrm{d}}$'




"""Unit tests for logout functionality."""

import pytest
from unittest.mock import Mock, patch, MagicMock
from idtap.client import SwaraClient
from idtap.secure_storage import SecureTokenStorage


class TestLogoutFunctionality:
    """Test cases for the logout method."""

    @pytest.mark.integration
    @patch('idtap.client.login_google')
    def test_logout_success(self, mock_login):
        """Test successful logout."""
        # Mock the authentication
        mock_login.return_value = {
            'id_token': 'fake_token',
            'profile': {'name': 'Test User', '_id': 'test123'}
        }
        
        # Create client and mock storage
        client = SwaraClient()
        client.token = 'fake_token'
        client.user = {'name': 'Test User', '_id': 'test123'}
        
        # Mock the secure storage clear_tokens method
        client.secure_storage.clear_tokens = Mock(return_value=True)
        
        # Test logout with confirmation
        with patch('builtins.input', return_value='yes'):
            result = client.logout()
        
        # Verify logout was successful
        assert result is True
        assert client.token is None
        assert client.user is None
        client.secure_storage.clear_tokens.assert_called_once()

    @pytest.mark.integration
    @patch('idtap.client.login_google')
    def test_logout_cancelled(self, mock_login):
        """Test logout cancellation."""
        # Mock the authentication
        mock_login.return_value = {
            'id_token': 'fake_token', 
            'profile': {'name': 'Test User', '_id': 'test123'}
        }
        
        client = SwaraClient()
        client.token = 'fake_token'
        client.user = {'name': 'Test User', '_id': 'test123'}
        
        # Mock storage
        client.secure_storage.clear_tokens = Mock(return_value=True)
        
        # Test logout cancellation
        with patch('builtins.input', return_value='no'):
            result = client.logout()
        
        # Verify logout was cancelled and state preserved
        assert result is False
        assert client.token == 'fake_token'
        assert client.user == {'name': 'Test User', '_id': 'test123'}
        client.secure_storage.clear_tokens.assert_not_called()

    @pytest.mark.integration
    @patch('idtap.client.login_google')
    def test_logout_programmatic(self, mock_login):
        """Test programmatic logout without confirmation."""
        # Mock the authentication
        mock_login.return_value = {
            'id_token': 'fake_token',
            'profile': {'name': 'Test User', '_id': 'test123'}
        }
        
        client = SwaraClient()
        client.token = 'fake_token'
        client.user = {'name': 'Test User', '_id': 'test123'}
        
        # Mock storage
        client.secure_storage.clear_tokens = Mock(return_value=True)
        
        # Test programmatic logout (no user input required)
        result = client.logout(confirm=True)
        
        # Verify logout was successful
        assert result is True
        assert client.token is None
        assert client.user is None
        client.secure_storage.clear_tokens.assert_called_once()

    @pytest.mark.integration
    @patch('idtap.client.login_google')
    def test_logout_storage_failure(self, mock_login):
        """Test logout when storage clearing fails."""
        # Mock the authentication
        mock_login.return_value = {
            'id_token': 'fake_token',
            'profile': {'name': 'Test User', '_id': 'test123'}
        }
        
        client = SwaraClient()
        client.token = 'fake_token'
        client.user = {'name': 'Test User', '_id': 'test123'}
        
        # Mock storage failure
        client.secure_storage.clear_tokens = Mock(return_value=False)
        
        # Test logout with storage failure
        result = client.logout(confirm=True)
        
        # Verify partial failure is handled
        assert result is False
        client.secure_storage.clear_tokens.assert_called_once()

    @pytest.mark.integration
    @patch('idtap.client.login_google')
    def test_logout_exception_handling(self, mock_login):
        """Test logout exception handling."""
        # Mock the authentication
        mock_login.return_value = {
            'id_token': 'fake_token',
            'profile': {'name': 'Test User', '_id': 'test123'}
        }
        
        client = SwaraClient()
        client.token = 'fake_token'
        client.user = {'name': 'Test User', '_id': 'test123'}
        
        # Mock storage to raise exception
        client.secure_storage.clear_tokens = Mock(side_effect=Exception("Storage error"))
        
        # Test logout with exception
        result = client.logout(confirm=True)
        
        # Verify exception is handled gracefully
        assert result is False
        client.secure_storage.clear_tokens.assert_called_once()


class TestSecureTokenStorage:
    """Test cases for SecureTokenStorage clear_tokens method."""
    
    def test_clear_tokens_method_exists(self):
        """Verify the clear_tokens method exists."""
        storage = SecureTokenStorage()
        assert hasattr(storage, 'clear_tokens')
        assert callable(getattr(storage, 'clear_tokens'))
    
    @patch('pathlib.Path.exists')
    @patch('pathlib.Path.unlink')
    def test_clear_tokens_removes_files(self, mock_unlink, mock_exists):
        """Test that clear_tokens removes all token files."""
        # Mock files exist
        mock_exists.return_value = True
        
        storage = SecureTokenStorage()
        
        # Mock keyring operations
        with patch('idtap.secure_storage.KEYRING_AVAILABLE', True), \
             patch('idtap.secure_storage.keyring.delete_password') as mock_delete:
            
            result = storage.clear_tokens()
            
            # Verify keyring deletion was attempted
            mock_delete.assert_called_once()
            
            # Verify file deletions were attempted
            assert mock_unlink.call_count >= 1  # Should try to delete encrypted and plaintext files
            assert result is True



"""Tests for spectrogram data access and visualization."""

import os
import sys
import gzip
import json
import pytest
import numpy as np
from pathlib import Path
from unittest.mock import Mock, patch, MagicMock

sys.path.insert(0, os.path.abspath('.'))

from idtap.spectrogram import SpectrogramData, SUPPORTED_COLORMAPS


# Helper function to create mock spectrogram data
def create_mock_spec_data(freq_bins=368, time_frames=1000):
    """Create mock spectrogram data for testing."""
    # Create synthetic spectrogram data (uint8)
    data = np.random.randint(0, 256, size=(freq_bins, time_frames), dtype=np.uint8)
    return data


def create_mock_compressed_data(freq_bins=368, time_frames=1000):
    """Create mock compressed spectrogram data."""
    data = create_mock_spec_data(freq_bins, time_frames)
    compressed = gzip.compress(data.tobytes())
    metadata = {"shape": [freq_bins, time_frames]}
    return compressed, metadata


class TestSpectrogramDataInit:
    """Test SpectrogramData initialization."""

    def test_init_with_valid_data(self):
        """Test initialization with valid numpy array."""
        data = create_mock_spec_data()
        spec = SpectrogramData(data, "test_audio_id")

        assert spec.audio_id == "test_audio_id"
        assert spec.shape == (368, 1000)
        assert spec.freq_range == (75.0, 2400.0)
        assert spec.bins_per_octave == 72

    def test_init_with_custom_params(self):
        """Test initialization with custom frequency range and bins."""
        data = create_mock_spec_data()
        spec = SpectrogramData(
            data, "test_id",
            freq_range=(100.0, 2000.0),
            bins_per_octave=60
        )

        assert spec.freq_range == (100.0, 2000.0)
        assert spec.bins_per_octave == 60

    def test_init_rejects_non_numpy(self):
        """Test that non-numpy array raises TypeError."""
        with pytest.raises(TypeError, match="data must be numpy array"):
            SpectrogramData([[1, 2], [3, 4]], "test_id")

    def test_init_rejects_non_uint8(self):
        """Test that non-uint8 dtype raises TypeError."""
        data = np.random.rand(100, 200).astype(np.float32)
        with pytest.raises(TypeError, match="data must be uint8 array"):
            SpectrogramData(data, "test_id")

    def test_init_rejects_non_2d(self):
        """Test that non-2D array raises ValueError."""
        data = np.random.randint(0, 256, size=(100,), dtype=np.uint8)
        with pytest.raises(ValueError, match="data must be 2D array"):
            SpectrogramData(data, "test_id")


class TestSpectrogramDataLoading:
    """Test loading spectrogram data from server."""

    @patch('idtap.client.SwaraClient')
    def test_from_audio_id_with_client(self, mock_client_class):
        """Test loading from audio_id with provided client."""
        # Setup mock client
        mock_client = Mock()
        compressed, metadata = create_mock_compressed_data()
        mock_client.download_spectrogram_data.return_value = compressed
        mock_client.download_spectrogram_metadata.return_value = metadata

        # Load spectrogram
        spec = SpectrogramData.from_audio_id("test_audio_id", mock_client)

        # Verify calls
        mock_client.download_spectrogram_data.assert_called_once_with("test_audio_id")
        mock_client.download_spectrogram_metadata.assert_called_once_with("test_audio_id")

        # Verify data
        assert spec.audio_id == "test_audio_id"
        assert spec.shape == (368, 1000)

    @patch('idtap.client.SwaraClient')
    def test_from_audio_id_creates_client(self, mock_client_class):
        """Test that from_audio_id creates client if not provided."""
        # Setup mock
        mock_client = Mock()
        compressed, metadata = create_mock_compressed_data()
        mock_client.download_spectrogram_data.return_value = compressed
        mock_client.download_spectrogram_metadata.return_value = metadata
        mock_client_class.return_value = mock_client

        # Load without providing client
        spec = SpectrogramData.from_audio_id("test_audio_id")

        # Verify client was created
        mock_client_class.assert_called_once()
        assert spec.audio_id == "test_audio_id"

    def test_from_piece_with_audio_id(self):
        """Test loading from Piece object with audio_id."""
        # Create mock piece
        mock_piece = Mock()
        mock_piece.audio_id = "piece_audio_id"

        # Create mock client
        mock_client = Mock()
        compressed, metadata = create_mock_compressed_data()
        mock_client.download_spectrogram_data.return_value = compressed
        mock_client.download_spectrogram_metadata.return_value = metadata

        # Load from piece
        spec = SpectrogramData.from_piece(mock_piece, mock_client)

        assert spec.audio_id == "piece_audio_id"

    def test_from_piece_without_audio_id(self):
        """Test from_piece returns None when no audio_id."""
        mock_piece = Mock()
        mock_piece.audio_id = None

        spec = SpectrogramData.from_piece(mock_piece)

        assert spec is None


class TestIntensityTransform:
    """Test intensity power transformation."""

    def test_apply_intensity_linear(self):
        """Test power=1.0 returns unchanged data."""
        data = create_mock_spec_data()
        spec = SpectrogramData(data, "test_id")

        result = spec.apply_intensity(power=1.0)

        np.testing.assert_array_equal(result, data)

    def test_apply_intensity_power_transform(self):
        """Test power transform increases contrast."""
        # Create test data with known values
        data = np.array([[100, 200]], dtype=np.uint8)
        spec = SpectrogramData(data, "test_id")

        result = spec.apply_intensity(power=2.0)

        # Higher power should make dark values darker, bright values relatively brighter
        assert result[0, 0] < data[0, 0]  # 100 gets darker
        assert result.dtype == np.uint8

    def test_apply_intensity_range_validation(self):
        """Test that power outside valid range raises ValueError."""
        data = create_mock_spec_data()
        spec = SpectrogramData(data, "test_id")

        with pytest.raises(ValueError, match="Power must be between 1.0 and 5.0"):
            spec.apply_intensity(power=0.5)

        with pytest.raises(ValueError, match="Power must be between 1.0 and 5.0"):
            spec.apply_intensity(power=6.0)

    def test_apply_intensity_preserves_shape(self):
        """Test that intensity transform preserves array shape."""
        data = create_mock_spec_data(100, 200)
        spec = SpectrogramData(data, "test_id")

        result = spec.apply_intensity(power=2.5)

        assert result.shape == data.shape


class TestColormapApplication:
    """Test colormap application."""

    def test_apply_colormap_default(self):
        """Test colormap with default viridis."""
        data = create_mock_spec_data(10, 20)
        spec = SpectrogramData(data, "test_id")

        rgb = spec.apply_colormap()

        assert rgb.shape == (10, 20, 3)
        assert rgb.dtype == np.uint8

    def test_apply_colormap_custom(self):
        """Test colormap with custom colormap name."""
        data = create_mock_spec_data(10, 20)
        spec = SpectrogramData(data, "test_id")

        rgb = spec.apply_colormap(cmap='plasma')

        assert rgb.shape == (10, 20, 3)

    def test_apply_colormap_invalid_name(self):
        """Test that invalid colormap name raises ValueError."""
        data = create_mock_spec_data()
        spec = SpectrogramData(data, "test_id")

        with pytest.raises(ValueError, match="Unknown colormap"):
            spec.apply_colormap(cmap='nonexistent_colormap')

    def test_apply_colormap_with_custom_data(self):
        """Test applying colormap to custom data array."""
        data = create_mock_spec_data()
        spec = SpectrogramData(data, "test_id")

        custom_data = np.ones((50, 100), dtype=np.uint8) * 128
        rgb = spec.apply_colormap(data=custom_data)

        assert rgb.shape == (50, 100, 3)


class TestCropping:
    """Test frequency and time cropping."""

    def test_crop_frequency(self):
        """Test frequency cropping."""
        data = create_mock_spec_data(368, 1000)
        spec = SpectrogramData(data, "test_id")

        # Crop to narrower frequency range
        cropped = spec.crop_frequency(min_hz=200, max_hz=800)

        # Should have fewer frequency bins
        assert cropped.shape[0] < spec.shape[0]
        assert cropped.shape[1] == spec.shape[1]  # Time unchanged
        assert cropped.freq_range[0] >= 200
        assert cropped.freq_range[1] <= 800

    def test_crop_time(self):
        """Test time cropping."""
        data = create_mock_spec_data(368, 1000)
        spec = SpectrogramData(data, "test_id")

        # Crop to 5-10 seconds
        cropped = spec.crop_time(start_time=5.0, end_time=10.0)

        # Should have fewer time frames
        assert cropped.shape[1] < spec.shape[1]
        assert cropped.shape[0] == spec.shape[0]  # Frequency unchanged

    def test_crop_chain(self):
        """Test chaining crop operations."""
        data = create_mock_spec_data(368, 1000)
        spec = SpectrogramData(data, "test_id")

        cropped = spec.crop_frequency(min_hz=200, max_hz=800).crop_time(start_time=2.0, end_time=8.0)

        assert cropped.shape[0] < spec.shape[0]
        assert cropped.shape[1] < spec.shape[1]


class TestProperties:
    """Test spectrogram properties."""

    def test_shape_property(self):
        """Test shape property."""
        data = create_mock_spec_data(100, 500)
        spec = SpectrogramData(data, "test_id")

        assert spec.shape == (100, 500)

    def test_duration_property(self):
        """Test duration calculation."""
        data = create_mock_spec_data(368, 1000)
        spec = SpectrogramData(data, "test_id")

        duration = spec.duration
        assert duration > 0
        assert isinstance(duration, float)

    def test_time_resolution_property(self):
        """Test time resolution property."""
        data = create_mock_spec_data()
        spec = SpectrogramData(data, "test_id")

        assert spec.time_resolution > 0
        assert spec.time_resolution < 0.1  # Should be in reasonable range

    def test_freq_bins_property(self):
        """Test frequency bins calculation."""
        data = create_mock_spec_data(368, 1000)
        spec = SpectrogramData(data, "test_id")

        freq_bins = spec.freq_bins

        assert len(freq_bins) == 368
        assert freq_bins[0] == spec.freq_range[0]  # Min frequency
        # Log spacing can slightly exceed max, allow 10% tolerance
        assert freq_bins[-1] <= spec.freq_range[1] * 1.1  # Max frequency (with tolerance)
        assert all(freq_bins[i] < freq_bins[i+1] for i in range(len(freq_bins)-1))  # Monotonic


class TestMatplotlibIntegration:
    """Test matplotlib integration methods."""

    def test_get_extent(self):
        """Test get_extent returns correct format."""
        data = create_mock_spec_data()
        spec = SpectrogramData(data, "test_id")

        extent = spec.get_extent()

        assert len(extent) == 4
        assert extent[0] == 0  # left (start time)
        assert extent[1] == spec.duration  # right (end time)
        assert extent[2] == spec.freq_range[0]  # bottom (min freq)
        assert extent[3] == spec.freq_range[1]  # top (max freq)

    def test_get_plot_data_without_colormap(self):
        """Test get_plot_data without colormap."""
        data = create_mock_spec_data(100, 200)
        spec = SpectrogramData(data, "test_id")

        plot_data, extent = spec.get_plot_data(power=2.0, apply_cmap=False)

        assert plot_data.shape == (100, 200)
        assert plot_data.dtype == np.uint8
        assert len(extent) == 4

    def test_get_plot_data_with_colormap(self):
        """Test get_plot_data with colormap applied."""
        data = create_mock_spec_data(100, 200)
        spec = SpectrogramData(data, "test_id")

        plot_data, extent = spec.get_plot_data(power=1.5, apply_cmap=True, cmap='plasma')

        assert plot_data.shape == (100, 200, 3)  # RGB
        assert plot_data.dtype == np.uint8

    @patch('matplotlib.pyplot.subplots')
    def test_plot_on_axis(self, mock_subplots):
        """Test plotting on matplotlib axis."""
        data = create_mock_spec_data()
        spec = SpectrogramData(data, "test_id")

        # Create mock axis
        mock_ax = MagicMock()
        mock_im = MagicMock()
        mock_ax.imshow.return_value = mock_im

        # Plot on axis
        im = spec.plot_on_axis(mock_ax, power=2.0, cmap='viridis', alpha=0.7)

        # Verify imshow was called
        mock_ax.imshow.assert_called_once()
        call_kwargs = mock_ax.imshow.call_args[1]
        assert call_kwargs['cmap'] == 'viridis'
        assert call_kwargs['alpha'] == 0.7
        assert call_kwargs['origin'] == 'lower'
        assert call_kwargs['aspect'] == 'auto'


class TestImageGeneration:
    """Test image generation methods."""

    def test_to_image_basic(self):
        """Test basic image generation."""
        data = create_mock_spec_data(100, 200)
        spec = SpectrogramData(data, "test_id")

        img = spec.to_image()

        assert img.mode == 'RGB'
        assert img.size == (200, 100)  # PIL uses (width, height)

    def test_to_image_with_resize(self):
        """Test image generation with resizing."""
        data = create_mock_spec_data(100, 200)
        spec = SpectrogramData(data, "test_id")

        img = spec.to_image(width=400, height=200)

        assert img.size == (400, 200)

    def test_to_image_with_power(self):
        """Test image generation with power transform."""
        data = create_mock_spec_data(100, 200)
        spec = SpectrogramData(data, "test_id")

        img = spec.to_image(power=2.5, cmap='plasma')

        assert img.mode == 'RGB'

    @patch('matplotlib.pyplot.colorbar')
    @patch('matplotlib.pyplot.subplots')
    def test_to_matplotlib(self, mock_subplots, mock_colorbar):
        """Test matplotlib figure generation."""
        data = create_mock_spec_data()
        spec = SpectrogramData(data, "test_id")

        # Setup mocks
        mock_fig = MagicMock()
        mock_ax = MagicMock()
        mock_im = MagicMock()
        mock_ax.imshow.return_value = mock_im
        mock_subplots.return_value = (mock_fig, mock_ax)

        fig = spec.to_matplotlib(power=2.0, cmap='viridis', show_colorbar=True)

        # Verify figure was created
        mock_subplots.assert_called_once()
        mock_ax.imshow.assert_called_once()
        mock_colorbar.assert_called_once()

    def test_save(self, tmp_path):
        """Test saving spectrogram to file."""
        data = create_mock_spec_data(50, 100)
        spec = SpectrogramData(data, "test_id")

        filepath = tmp_path / "test_spec.png"
        spec.save(str(filepath), power=1.5, cmap='viridis')

        assert filepath.exists()


class TestConstants:
    """Test module-level constants."""

    def test_supported_colormaps(self):
        """Test that SUPPORTED_COLORMAPS list is defined."""
        assert isinstance(SUPPORTED_COLORMAPS, list)
        assert len(SUPPORTED_COLORMAPS) > 0
        assert 'viridis' in SUPPORTED_COLORMAPS
        assert 'plasma' in SUPPORTED_COLORMAPS

    def test_default_freq_range(self):
        """Test default frequency range constant."""
        assert SpectrogramData.DEFAULT_FREQ_RANGE == (75.0, 2400.0)

    def test_default_bins_per_octave(self):
        """Test default bins per octave constant."""
        assert SpectrogramData.DEFAULT_BINS_PER_OCTAVE == 72



"""Query types for musical transcription analysis.

This module defines Python equivalents of TypeScript query types to ensure
exact data structure compatibility between the web app and Python API.
"""

from __future__ import annotations
from typing import Dict, List, Optional, Union, Literal, TypedDict
from enum import Enum
from dataclasses import dataclass

from .classes.trajectory import Trajectory
from .classes.pitch import Pitch


class CategoryType(str, Enum):
    """Query category types for different musical elements."""
    TRAJECTORY_ID = "trajectoryID"
    PITCH = "pitch"
    VOWEL = "vowel"
    STARTING_CONSONANT = "startingConsonant"
    ENDING_CONSONANT = "endingConsonant"
    ANY_CONSONANT = "anyConsonant"
    PITCH_SEQUENCE_STRICT = "pitchSequenceStrict"
    PITCH_SEQUENCE_LOOSE = "pitchSequenceLoose"
    TRAJ_SEQUENCE_STRICT = "trajSequenceStrict"
    TRAJ_SEQUENCE_LOOSE = "trajSequenceLoose"
    SECTION_TOP_LEVEL = "sectionTopLevel"
    ALAP_SECTION = "alapSection"
    COMP_TYPE = "compType"
    COMP_SEC_TEMPO = "compSecTempo"
    TALA = "tala"
    PHRASE_TYPE = "phraseType"
    ELABORATION_TYPE = "elaborationType"
    VOCAL_ART_TYPE = "vocalArtType"
    INST_ART_TYPE = "instArtType"
    INCIDENTAL = "incidental"


class DesignatorType(str, Enum):
    """Query designator types for matching behavior."""
    INCLUDES = "includes"
    EXCLUDES = "excludes"
    STARTS_WITH = "startsWith"
    ENDS_WITH = "endsWith"


class SegmentationType(str, Enum):
    """Query segmentation types for different analysis units."""
    PHRASE = "phrase"
    GROUP = "group"
    SEQUENCE_OF_TRAJECTORIES = "sequenceOfTrajectories"
    CONNECTED_SEQUENCE_OF_TRAJECTORIES = "connectedSequenceOfTrajectories"


# Section categorization types (matching TypeScript SecCatType)
class SecCatType(TypedDict, total=False):
    """Section categorization structure."""
    # Pre-Chiz Alap section
    pre_chiz_alap: Dict[str, bool]
    
    # Alap section types
    alap: Dict[str, bool]
    
    # Composition types
    composition_type: Dict[str, bool]
    
    # Tempo/section types
    comp_section_tempo: Dict[str, bool]
    
    # Tala types
    tala: Dict[str, bool]
    
    # Other categories
    improvisation: Dict[str, bool]
    other: Dict[str, bool]
    
    # Top level category
    top_level: Literal[
        "Pre-Chiz Alap", 
        "Alap", 
        "Composition", 
        "Improvisation", 
        "Other", 
        "None"
    ]


# Phrase categorization types (matching TypeScript PhraseCatType)
class PhraseCatType(TypedDict, total=False):
    """Phrase categorization structure."""
    phrase: Dict[str, bool]
    elaboration: Dict[str, bool]
    vocal_articulation: Dict[str, bool]
    instrumental_articulation: Dict[str, bool]
    incidental: Dict[str, bool]


class QueryType(TypedDict, total=False):
    """Single query specification."""
    category: CategoryType
    designator: DesignatorType
    pitch: Optional[Pitch]
    trajectory_id: Optional[int]
    vowel: Optional[str]
    consonant: Optional[str]
    pitch_sequence: Optional[List[Pitch]]
    traj_id_sequence: Optional[List[int]]
    section_top_level: Optional[str]
    alap_section: Optional[str]
    comp_type: Optional[str]
    comp_sec_tempo: Optional[str]
    tala: Optional[str]
    phrase_type: Optional[str]
    elaboration_type: Optional[str]
    vocal_art_type: Optional[str]
    inst_art_type: Optional[str]
    incidental: Optional[str]
    instrument_idx: int


@dataclass
class QueryAnswerType:
    """Result of a query execution."""
    trajectories: List[Trajectory]
    identifier: Union[int, str, Dict[str, int]]  # Can be phraseIdx or {phraseIdx, trajIdx}
    title: str
    start_time: float
    end_time: float
    duration: float
    segmentation: SegmentationType
    
    def to_json(self) -> Dict:
        """Convert to JSON-serializable dictionary with camelCase keys."""
        from .utils import to_camel_case
        return to_camel_case({
            "trajectories": [traj.to_json() for traj in self.trajectories],
            "identifier": self.identifier,
            "title": self.title,
            "start_time": self.start_time,
            "end_time": self.end_time,
            "duration": self.duration,
            "segmentation": self.segmentation.value if isinstance(self.segmentation, SegmentationType) else self.segmentation
        })
    
    @staticmethod
    def from_json(data: Dict) -> QueryAnswerType:
        """Create from JSON dictionary with camelCase keys."""
        from .utils import to_snake_case
        snake_data = to_snake_case(data)
        
        # Import here to avoid circular imports
        from .classes.trajectory import Trajectory
        
        trajectories = [Trajectory.from_json(t) for t in snake_data["trajectories"]]
        
        return QueryAnswerType(
            trajectories=trajectories,
            identifier=snake_data["identifier"],
            title=snake_data["title"],
            start_time=snake_data["start_time"],
            end_time=snake_data["end_time"],
            duration=snake_data["duration"],
            segmentation=SegmentationType(snake_data["segmentation"])
        )


# Type aliases for multi-query returns
MultipleReturnType = tuple[
    List[List[Trajectory]],  # trajectories
    List[Union[int, str, Dict[str, int]]],  # identifiers
    List[QueryAnswerType]  # query answers
]


class MultipleOptionType(TypedDict, total=False):
    """Options for multiple query execution."""
    transcription_id: Optional[str]
    piece: Optional[object]  # Will be Piece type, but avoiding circular import
    segmentation: SegmentationType
    sequence_length: Optional[int]
    min_dur: float
    max_dur: float
    every: bool  # True = all queries must match, False = any query can match
    instrument_idx: int


# Default categorization structures
def init_sec_categorization() -> SecCatType:
    """Initialize default section categorization structure."""
    return {
        "pre_chiz_alap": {"Pre-Chiz Alap": False},
        "alap": {
            "Alap": False,
            "Jor": False,
            "Alap-Jhala": False
        },
        "composition_type": {
            "Dhrupad": False,
            "Bandish": False,
            "Thumri": False,
            "Ghazal": False,
            "Qawwali": False,
            "Dhun": False,
            "Tappa": False,
            "Bhajan": False,
            "Kirtan": False,
            "Kriti": False,
            "Masitkhani Gat": False,
            "Razakhani Gat": False,
            "Ferozkhani Gat": False,
        },
        "comp_section_tempo": {
            "Ati Vilambit": False,
            "Vilambit": False,
            "Madhya": False,
            "Drut": False,
            "Ati Drut": False,
            "Jhala": False,
        },
        "tala": {
            "Ektal": False,
            "Tintal": False,
            "Rupak": False
        },
        "improvisation": {"Improvisation": False},
        "other": {"Other": False},
        "top_level": "None"
    }


def init_phrase_categorization() -> PhraseCatType:
    """Initialize default phrase categorization structure."""
    return {
        "phrase": {
            "Mohra": False,
            "Mukra": False,
            "Asthai": False,
            "Antara": False,
            "Manjha": False,
            "Abhog": False,
            "Sanchari": False,
            "Jhala": False
        },
        "elaboration": {
            "Vistar": False,
            "Barhat": False,
            "Prastar": False,
            "Bol Banao": False,
            "Bol Alap": False,
            "Bol Bandt": False,
            "Behlava": False,
            "Gat-kari": False,
            "Tan (Sapat)": False,
            "Tan (Gamak)": False,
            "Laykari": False,
            "Tihai": False,
            "Chakradar": False,
        },
        "vocal_articulation": {
            "Bol": False,
            "Non-Tom": False,
            "Tarana": False,
            "Aakar": False,
            "Sargam": False
        },
        "instrumental_articulation": {
            "Bol": False,
            "Non-Bol": False
        },
        "incidental": {
            "Talk/Conversation": False,
            "Praise ('Vah')": False,
            "Tuning": False,
            "Pause": False,
        }
    }


import os
import sys

sys.path.insert(0, os.path.abspath('.'))

from idtap.classes.section import Section
from idtap.classes.phrase import Phrase
from idtap.classes.trajectory import Trajectory
from idtap.classes.pitch import Pitch


# Tests mirror src/ts/tests/section.test.ts


def test_section_aggregates():
    p1 = Phrase({'trajectories': [Trajectory()]})
    p2 = Phrase({'trajectories': [Trajectory()]})
    sec = Section({'phrases': [p1, p2]})
    assert len(sec.trajectories) == 2
    assert len(sec.all_pitches()) == 2


def test_section_all_pitches_and_trajectories_getters():
    sa1 = Trajectory({'pitches': [Pitch({'swara': 'sa'})]})
    sa2 = Trajectory({'pitches': [Pitch({'swara': 'sa'})]})
    re = Trajectory({'pitches': [Pitch({'swara': 're'})]})

    p1 = Phrase({'trajectories': [sa1, sa2]})
    p2 = Phrase({'trajectories': [re]})
    sec = Section({'phrases': [p1, p2]})

    assert len(sec.all_pitches(False)) == 2
    assert sec.trajectories == [sa1, sa2, re]




from __future__ import annotations
from typing import List, Dict, Optional, Union, Literal, TYPE_CHECKING
from uuid import uuid4
import math

if TYPE_CHECKING:
    from .musical_time import MusicalTime

# Tempo validation constants
MIN_TEMPO_BPM = 20  # Very slow musical pieces (e.g., some alap sections)
MAX_TEMPO_BPM = 300  # Very fast musical pieces



def find_closest_idxs(trials: List[float], items: List[float]) -> List[int]:
    """Return indexes of items closest to each trial (greedy)."""
    used: set[int] = set()
    out: List[int] = []
    for trial in trials:
        diffs = [(abs(trial - item), idx) for idx, item in enumerate(items)
                 if idx not in used]
        diffs.sort(key=lambda x: x[0])
        if not diffs:
            raise ValueError("not enough items to match trials")
        used.add(diffs[0][1])
        out.append(diffs[0][1])
    return out


class Pulse:
    def __init__(self, real_time: float = 0.0, unique_id: Optional[str] = None,
                 affiliations: Optional[List[Dict]] = None,
                 meter_id: Optional[str] = None,
                 corporeal: bool = True) -> None:
        # Parameter validation
        self._validate_parameters({'real_time': real_time, 'unique_id': unique_id, 
                                  'affiliations': affiliations, 'meter_id': meter_id, 'corporeal': corporeal})
        self.real_time = real_time
        self.unique_id = unique_id or str(uuid4())
        self.affiliations: List[Dict] = affiliations or []
        self.meter_id = meter_id
        self.corporeal = corporeal

    def _validate_parameters(self, opts: Dict) -> None:
        """Validate constructor parameters and provide helpful error messages."""
        if not isinstance(opts.get('real_time', 0.0), (int, float)):
            raise TypeError(f"Parameter 'real_time' must be a number, got {type(opts['real_time']).__name__}")
        
        if opts.get('real_time', 0.0) < 0:
            raise ValueError(f"Parameter 'real_time' must be non-negative, got {opts['real_time']}")
        
        if 'unique_id' in opts and opts['unique_id'] is not None and not isinstance(opts['unique_id'], str):
            raise TypeError(f"Parameter 'unique_id' must be a string, got {type(opts['unique_id']).__name__}")
        
        if 'affiliations' in opts and opts['affiliations'] is not None:
            if not isinstance(opts['affiliations'], list):
                raise TypeError(f"Parameter 'affiliations' must be a list, got {type(opts['affiliations']).__name__}")
            if not all(isinstance(item, dict) for item in opts['affiliations']):
                raise TypeError("All items in 'affiliations' must be dictionaries")
        
        if 'meter_id' in opts and opts['meter_id'] is not None and not isinstance(opts['meter_id'], str):
            raise TypeError(f"Parameter 'meter_id' must be a string, got {type(opts['meter_id']).__name__}")
        
        if not isinstance(opts.get('corporeal', True), bool):
            raise TypeError(f"Parameter 'corporeal' must be a boolean, got {type(opts['corporeal']).__name__}")

    @staticmethod
    def from_json(obj: Dict) -> 'Pulse':
        return Pulse(
            real_time=obj.get('realTime', 0.0),
            unique_id=obj.get('uniqueId'),
            affiliations=obj.get('affiliations'),
            meter_id=obj.get('meterId'),
            corporeal=obj.get('corporeal', True),
        )

    def to_json(self) -> Dict:
        return {
            'realTime': self.real_time,
            'uniqueId': self.unique_id,
            'affiliations': self.affiliations,
            'meterId': self.meter_id,
            'corporeal': self.corporeal,
        }

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Pulse) and self.to_json() == other.to_json()


class PulseStructure:
    def __init__(self, tempo: float = 60.0, size: int = 4,
                 start_time: float = 0.0, unique_id: Optional[str] = None,
                 front_weighted: bool = True, layer: Optional[int] = None,
                 parent_pulse_id: Optional[str] = None,
                 primary: bool = True, segmented_meter_idx: int = 0,
                 meter_id: Optional[str] = None,
                 pulses: Optional[List[Pulse | Dict]] = None) -> None:
        # Parameter validation
        self._validate_parameters({
            'tempo': tempo, 'size': size, 'start_time': start_time, 'unique_id': unique_id,
            'front_weighted': front_weighted, 'layer': layer, 'parent_pulse_id': parent_pulse_id,
            'primary': primary, 'segmented_meter_idx': segmented_meter_idx, 'meter_id': meter_id,
            'pulses': pulses
        })
        self.tempo = tempo
        self.pulse_dur = 60.0 / tempo
        self.size = size
        self.start_time = start_time
        self.unique_id = unique_id or str(uuid4())
        self.front_weighted = front_weighted
        self.layer = layer
        self.parent_pulse_id = parent_pulse_id
        self.primary = primary
        self.segmented_meter_idx = segmented_meter_idx
        self.meter_id = meter_id

        if pulses is not None:
            self.pulses = [p if isinstance(p, Pulse) else Pulse.from_json(p)
                           for p in pulses]
        else:
            self.pulses = [
                Pulse(
                    real_time=start_time + i * self.pulse_dur,
                    affiliations=[{
                        'psId': self.unique_id,
                        'idx': i,
                        'layer': self.layer,
                        'segmentedMeterIdx': self.segmented_meter_idx,
                        'strong': (i == 0) if front_weighted else (i == size - 1),
                    }],
                    meter_id=meter_id
                ) for i in range(size)
            ]

    def _validate_parameters(self, opts: Dict) -> None:
        """Validate constructor parameters and provide helpful error messages."""
        if not isinstance(opts.get('tempo', 60.0), (int, float)):
            raise TypeError(f"Parameter 'tempo' must be a number, got {type(opts['tempo']).__name__}")
        
        if opts.get('tempo', 60.0) <= 0:
            raise ValueError(f"Parameter 'tempo' must be positive, got {opts['tempo']}")
        
        if opts.get('tempo', 60.0) < MIN_TEMPO_BPM or opts.get('tempo', 60.0) > MAX_TEMPO_BPM:
            import warnings
            warnings.warn(f"Tempo {opts['tempo']} BPM is outside typical range ({MIN_TEMPO_BPM}-{MAX_TEMPO_BPM} BPM)", UserWarning)
        
        if not isinstance(opts.get('size', 4), int):
            raise TypeError(f"Parameter 'size' must be an integer, got {type(opts['size']).__name__}")
        
        if opts.get('size', 4) <= 0:
            raise ValueError(f"Parameter 'size' must be positive, got {opts['size']}")
        
        if not isinstance(opts.get('start_time', 0.0), (int, float)):
            raise TypeError(f"Parameter 'start_time' must be a number, got {type(opts['start_time']).__name__}")
        
        if opts.get('start_time', 0.0) < 0:
            raise ValueError(f"Parameter 'start_time' must be non-negative, got {opts['start_time']}")
        
        if 'unique_id' in opts and opts['unique_id'] is not None and not isinstance(opts['unique_id'], str):
            raise TypeError(f"Parameter 'unique_id' must be a string, got {type(opts['unique_id']).__name__}")
        
        if not isinstance(opts.get('front_weighted', True), bool):
            raise TypeError(f"Parameter 'front_weighted' must be a boolean, got {type(opts['front_weighted']).__name__}")
        
        if 'layer' in opts and opts['layer'] is not None:
            if not isinstance(opts['layer'], int):
                raise TypeError(f"Parameter 'layer' must be an integer, got {type(opts['layer']).__name__}")
            if opts['layer'] < 0:
                raise ValueError(f"Parameter 'layer' must be non-negative, got {opts['layer']}")
        
        if 'parent_pulse_id' in opts and opts['parent_pulse_id'] is not None and not isinstance(opts['parent_pulse_id'], str):
            raise TypeError(f"Parameter 'parent_pulse_id' must be a string, got {type(opts['parent_pulse_id']).__name__}")
        
        if not isinstance(opts.get('primary', True), bool):
            raise TypeError(f"Parameter 'primary' must be a boolean, got {type(opts['primary']).__name__}")
        
        if not isinstance(opts.get('segmented_meter_idx', 0), int):
            raise TypeError(f"Parameter 'segmented_meter_idx' must be an integer, got {type(opts['segmented_meter_idx']).__name__}")
        
        if opts.get('segmented_meter_idx', 0) < 0:
            raise ValueError(f"Parameter 'segmented_meter_idx' must be non-negative, got {opts['segmented_meter_idx']}")
        
        if 'meter_id' in opts and opts['meter_id'] is not None and not isinstance(opts['meter_id'], str):
            raise TypeError(f"Parameter 'meter_id' must be a string, got {type(opts['meter_id']).__name__}")
        
        if 'pulses' in opts and opts['pulses'] is not None:
            if not isinstance(opts['pulses'], list):
                raise TypeError(f"Parameter 'pulses' must be a list, got {type(opts['pulses']).__name__}")

    @property
    def dur_tot(self) -> float:
        return self.size * self.pulse_dur

    def set_tempo(self, new_tempo: float) -> None:
        self.tempo = new_tempo
        self.pulse_dur = 60.0 / new_tempo
        for i, pulse in enumerate(self.pulses):
            pulse.real_time = self.start_time + i * self.pulse_dur

    def set_start_time(self, new_start: float) -> None:
        diff = new_start - self.start_time
        self.start_time = new_start
        for pulse in self.pulses:
            pulse.real_time += diff

    @staticmethod
    def from_pulse(pulse: Pulse, duration: float, size: int,
                   front_weighted: bool = True, layer: int = 0) -> 'PulseStructure':
        tempo = 60 * size / duration
        ps = PulseStructure(tempo=tempo, size=size, start_time=pulse.real_time,
                            front_weighted=front_weighted, layer=layer,
                            parent_pulse_id=pulse.unique_id, meter_id=pulse.meter_id)
        idx = 0 if front_weighted else ps.size - 1
        pulse.affiliations.append({
            'psId': ps.unique_id,
            'idx': idx,
            'layer': layer,
            'segmentedMeterIdx': 0,
            'strong': True,
        })
        ps.pulses[idx] = pulse
        return ps

    def to_json(self) -> Dict:
        return {
            'pulses': [p.to_json() for p in self.pulses],
            'tempo': self.tempo,
            'pulseDur': self.pulse_dur,
            'size': self.size,
            'startTime': self.start_time,
            'uniqueId': self.unique_id,
            'frontWeighted': self.front_weighted,
            'layer': self.layer,
            'parentPulseID': self.parent_pulse_id,
            'primary': self.primary,
            'segmentedMeterIdx': self.segmented_meter_idx,
            'meterId': self.meter_id,
            'offsets': [0.0] * self.size,
        }

    @staticmethod
    def from_json(obj: Dict) -> 'PulseStructure':
        return PulseStructure(
            tempo=obj.get('tempo', 60.0),
            size=obj.get('size', 4),
            start_time=obj.get('startTime', 0.0),
            unique_id=obj.get('uniqueId'),
            front_weighted=obj.get('frontWeighted', True),
            layer=obj.get('layer'),
            parent_pulse_id=obj.get('parentPulseID'),
            primary=obj.get('primary', True),
            segmented_meter_idx=obj.get('segmentedMeterIdx', 0),
            meter_id=obj.get('meterId'),
            pulses=[Pulse.from_json(p) for p in obj.get('pulses', [])]
        )

    def __eq__(self, other: object) -> bool:
        return isinstance(other, PulseStructure) and self.to_json() == other.to_json()


class Meter:
    def __init__(self, hierarchy: Optional[List[int | List[int]]] = None,
                 start_time: float = 0.0, tempo: float = 60.0,
                 unique_id: Optional[str] = None, repetitions: int = 1) -> None:
        # Parameter validation
        self._validate_parameters({
            'hierarchy': hierarchy, 'start_time': start_time, 'tempo': tempo,
            'unique_id': unique_id, 'repetitions': repetitions
        })
        self.hierarchy = hierarchy or [4, 4]
        self.start_time = start_time
        self.tempo = tempo
        self.unique_id = unique_id or str(uuid4())
        self.repetitions = repetitions
        self.pulse_structures: List[List[PulseStructure]] = []
        self._generate_pulse_structures()

    def _validate_parameters(self, opts: Dict) -> None:
        """Validate constructor parameters and provide helpful error messages."""
        if 'hierarchy' in opts and opts['hierarchy'] is not None:
            if not isinstance(opts['hierarchy'], list):
                raise TypeError(f"Parameter 'hierarchy' must be a list, got {type(opts['hierarchy']).__name__}")
            
            if len(opts['hierarchy']) == 0:
                raise ValueError("Parameter 'hierarchy' cannot be empty")
            
            for i, level in enumerate(opts['hierarchy']):
                if isinstance(level, list):
                    if not all(isinstance(item, int) for item in level):
                        raise TypeError(f"All items in hierarchy[{i}] must be integers")
                    if any(item <= 0 for item in level):
                        raise ValueError(f"All items in hierarchy[{i}] must be positive")
                elif isinstance(level, int):
                    if level <= 0:
                        raise ValueError(f"hierarchy[{i}] must be positive, got {level}")
                else:
                    raise TypeError(f"hierarchy[{i}] must be an integer or list of integers, got {type(level).__name__}")
        
        if not isinstance(opts.get('start_time', 0.0), (int, float)):
            raise TypeError(f"Parameter 'start_time' must be a number, got {type(opts['start_time']).__name__}")
        
        if opts.get('start_time', 0.0) < 0:
            raise ValueError(f"Parameter 'start_time' must be non-negative, got {opts['start_time']}")
        
        if not isinstance(opts.get('tempo', 60.0), (int, float)):
            raise TypeError(f"Parameter 'tempo' must be a number, got {type(opts['tempo']).__name__}")
        
        if opts.get('tempo', 60.0) <= 0:
            raise ValueError(f"Parameter 'tempo' must be positive, got {opts['tempo']}")
        
        if opts.get('tempo', 60.0) < MIN_TEMPO_BPM or opts.get('tempo', 60.0) > MAX_TEMPO_BPM:
            import warnings
            warnings.warn(f"Tempo {opts['tempo']} BPM is outside typical range ({MIN_TEMPO_BPM}-{MAX_TEMPO_BPM} BPM)", UserWarning)
        
        if 'unique_id' in opts and opts['unique_id'] is not None and not isinstance(opts['unique_id'], str):
            raise TypeError(f"Parameter 'unique_id' must be a string, got {type(opts['unique_id']).__name__}")
        
        if not isinstance(opts.get('repetitions', 1), int):
            raise TypeError(f"Parameter 'repetitions' must be an integer, got {type(opts['repetitions']).__name__}")
        
        if opts.get('repetitions', 1) <= 0:
            raise ValueError(f"Parameter 'repetitions' must be positive, got {opts['repetitions']}")

    # helper values
    @property
    def _top_size(self) -> int:
        h0 = self.hierarchy[0]
        return sum(h0) if isinstance(h0, list) else int(h0)

    @property
    def _bottom_mult(self) -> int:
        mult = 1
        for h in self.hierarchy[1:]:
            mult *= int(h)
        return mult

    @property
    def _pulses_per_cycle(self) -> int:
        return self._top_size * self._bottom_mult

    @property
    def _pulse_dur(self) -> float:
        return 60.0 / self.tempo / self._bottom_mult

    @property
    def cycle_dur(self) -> float:
        return self._pulse_dur * self._pulses_per_cycle

    def _get_hierarchy_mult(self, layer: int) -> int:
        """Get the multiplier for a given hierarchy layer.

        Handles both simple numbers and complex arrays like [3, 2] -> 5

        Args:
            layer: The hierarchy layer index

        Returns:
            The multiplier for that layer (sum if array, value if int)
        """
        if layer >= len(self.hierarchy):
            return 1
        h = self.hierarchy[layer]
        if isinstance(h, int):
            return h
        else:
            return sum(h)

    @property
    def display_tempo(self) -> float:
        """Get the tempo as displayed in performance practice (at matra/beat level).

        In Hindustani music, the internal tempo is stored at the finest pulse level,
        but musicians typically think of tempo at layer 1 (the matra/beat level).

        For a hierarchy like [[4,4,4,4], 4] (Tintal) with internal tempo 60 BPM:
        - Layer 1 multiplier = 4
        - Display tempo = 60 * 4 = 240 BPM

        For complex hierarchies where layer 1 is an array like [3, 2],
        sum the elements to get the multiplier (5).

        Returns:
            The tempo at the matra/beat level (layer 1)
        """
        if len(self.hierarchy) < 2:
            return self.tempo
        return self.tempo * self._get_hierarchy_mult(1)

    @display_tempo.setter
    def display_tempo(self, new_tempo: float) -> None:
        """Set the tempo using the performance practice tempo (at matra/beat level).

        Converts the matra-level tempo back to internal pulse tempo.

        Args:
            new_tempo: The desired tempo at the matra/beat level
        """
        if len(self.hierarchy) < 2:
            # For single-layer hierarchies, display tempo equals internal tempo
            self.tempo = new_tempo
            self._generate_pulse_structures()
        else:
            internal_tempo = new_tempo / self._get_hierarchy_mult(1)
            self.tempo = internal_tempo
            self._generate_pulse_structures()

    def get_tempo_at_layer(self, layer: int) -> float:
        """Get the tempo at a specific hierarchical layer.

        Args:
            layer: The hierarchy layer (0 = coarsest/vibhag, higher = finer subdivisions)

        Returns:
            The tempo (BPM) at that layer

        Raises:
            ValueError: If layer is out of bounds
        """
        if layer < 0 or layer >= len(self.hierarchy):
            raise ValueError(f"Layer {layer} is out of bounds for hierarchy with {len(self.hierarchy)} layers")

        # Start with base tempo and multiply by each layer's subdivision
        result_tempo = self.tempo
        for i in range(1, layer + 1):
            result_tempo *= self._get_hierarchy_mult(i)
        return result_tempo

    def _generate_pulse_structures(self) -> None:
        self.pulse_structures = [[]]
        # single layer of pulses for simplified implementation
        pulses: List[Pulse] = []
        for rep in range(self.repetitions):
            start = self.start_time + rep * self.cycle_dur
            for i in range(self._pulses_per_cycle):
                pulses.append(Pulse(real_time=start + i * self._pulse_dur,
                                    meter_id=self.unique_id))
        self.pulse_structures[0] = [PulseStructure(
            tempo=self.tempo,
            size=self._pulses_per_cycle,
            start_time=self.start_time,
            meter_id=self.unique_id,
            pulses=pulses,
        )]

    @property
    def all_pulses(self) -> List[Pulse]:
        """Get all pulses from the finest layer (lowest level) of the hierarchy.
        
        This concatenates pulses from all pulse structures in the last layer,
        matching the TypeScript implementation: lastLayer.map(ps => ps.pulses).flat()
        """
        if not self.pulse_structures or not self.pulse_structures[-1]:
            return []
        # Flatten all pulses from all structures in the finest layer
        return [pulse for ps in self.pulse_structures[-1] for pulse in ps.pulses]

    @property
    def real_times(self) -> List[float]:
        return [p.real_time for p in self.all_pulses]

    def offset_pulse(self, pulse: Pulse, offset: float) -> None:
        pulse.real_time += offset

    def reset_tempo(self) -> None:
        base = self.all_pulses[:self._pulses_per_cycle]
        diff = base[-1].real_time - base[0].real_time
        if len(base) > 1:
            bit = diff / (len(base) - 1)
            if bit > 0:
                self.tempo = 60.0 / (bit * self._bottom_mult)
        # pulse duration will be derived from tempo

    def grow_cycle(self) -> None:
        self.reset_tempo()
        start = self.start_time + self.repetitions * self.cycle_dur
        for i in range(self._pulses_per_cycle):
            new_pulse = Pulse(real_time=start + i * self._pulse_dur,
                              meter_id=self.unique_id)
            self.pulse_structures[0][0].pulses.append(new_pulse)
        self.repetitions += 1

    def add_time_points(self, time_points: List[float], layer: int = 1) -> None:
        time_points = sorted(time_points)
        for tp in time_points:
            self.pulse_structures[0][0].pulses.append(Pulse(real_time=tp,
                                                            meter_id=self.unique_id))
        self.pulse_structures[0][0].pulses.sort(key=lambda p: p.real_time)

    @staticmethod
    def from_time_points(time_points: List[float], hierarchy: List[Union[int, List[int]]], 
                        repetitions: int = 1, layer: int = 0) -> 'Meter':
        """Create a Meter from actual pulse time points, handling timing variations.
        
        This method creates a meter that accurately represents actual pulse timing
        (including rubato and tempo variations) rather than theoretical even spacing.
        Uses timing regularization algorithm to handle extreme deviations.
        
        Args:
            time_points: List of actual pulse times in seconds
            hierarchy: Meter hierarchy (e.g., [4, 4, 2])  
            repetitions: Number of cycle repetitions
            layer: Which hierarchical layer the time points represent (0 or 1)
            
        Returns:
            Meter object with pulses positioned at the provided time points
        """
        if not time_points or len(time_points) < 2:
            raise ValueError("Must provide at least two time points")
        
        if not hierarchy or len(hierarchy) < 1:
            raise ValueError("Must provide hierarchy to create Meter")
        
        # Work on a copy to avoid modifying the original
        time_points = sorted(time_points.copy())
        
        # Step 1: Timing regularization algorithm (from TypeScript)
        # Calculate pulse duration and handle extreme deviations
        diffs = [time_points[i+1] - time_points[i] for i in range(len(time_points) - 1)]
        pulse_dur = sum(diffs) / len(diffs)
        
        # Normalize deviations relative to pulse duration
        zeroed_tps = [tp - time_points[0] for tp in time_points]
        norms = [pulse_dur * i for i in range(len(time_points))]
        tp_diffs = [(zeroed_tps[i] - norms[i]) / pulse_dur for i in range(len(time_points))]
        
        # Insert intermediate time points when deviations exceed 40%
        max_iterations = 100  # Prevent infinite loops
        iteration = 0
        while any(abs(d) > 0.4 for d in tp_diffs) and iteration < max_iterations:
            abs_tp_diffs = [abs(d) for d in tp_diffs]
            biggest_idx = abs_tp_diffs.index(max(abs_tp_diffs))
            diff = tp_diffs[biggest_idx]
            
            if diff > 0:
                # Insert time point before the problematic one
                if biggest_idx > 0:
                    new_tp = (time_points[biggest_idx-1] + time_points[biggest_idx]) / 2
                    time_points.insert(biggest_idx, new_tp)
                else:
                    # Can't insert before first point, adjust differently
                    break
            else:
                # Insert time point after the problematic one
                if biggest_idx < len(time_points) - 1:
                    new_tp = (time_points[biggest_idx] + time_points[biggest_idx+1]) / 2
                    time_points.insert(biggest_idx+1, new_tp)
                else:
                    # Can't insert after last point, adjust differently
                    break
            
            # Recalculate deviations
            diffs = [time_points[i+1] - time_points[i] for i in range(len(time_points) - 1)]
            pulse_dur = sum(diffs) / len(diffs)
            zeroed_tps = [tp - time_points[0] for tp in time_points]
            norms = [pulse_dur * i for i in range(len(time_points))]
            tp_diffs = [(zeroed_tps[i] - norms[i]) / pulse_dur for i in range(len(time_points))]
            
            iteration += 1
        
        # Step 2: Calculate meter properties
        tempo = 60.0 / pulse_dur
        start_time = time_points[0]
        
        # Determine how many repetitions we need based on time points
        if isinstance(hierarchy[0], list):
            layer0_size = sum(hierarchy[0])
        else:
            layer0_size = hierarchy[0]
            
        # Calculate minimum repetitions needed
        min_reps = max(repetitions, (len(time_points) + layer0_size - 1) // layer0_size)
        
        # Step 3: Create theoretical meter
        meter = Meter(
            hierarchy=hierarchy,
            start_time=start_time,
            tempo=tempo,
            repetitions=min_reps
        )
        
        # Step 4: Adjust pulses to match actual time points
        finest_pulses = meter.all_pulses
        
        # Update pulse times to match provided time points
        for i, time_point in enumerate(time_points):
            if i < len(finest_pulses):
                finest_pulses[i].real_time = time_point
        
        # If we have fewer time points than pulses, extrapolate the remaining
        if len(time_points) < len(finest_pulses):
            # Use the calculated pulse duration to extrapolate
            last_provided_time = time_points[-1]
            for i in range(len(time_points), len(finest_pulses)):
                extrapolated_time = last_provided_time + (i - len(time_points) + 1) * pulse_dur
                finest_pulses[i].real_time = extrapolated_time
        
        return meter

    @staticmethod  
    def from_json(obj: Dict) -> 'Meter':
        m = Meter(hierarchy=obj.get('hierarchy'),
                  start_time=obj.get('startTime', 0.0),
                  tempo=obj.get('tempo', 60.0),
                  unique_id=obj.get('uniqueId'),
                  repetitions=obj.get('repetitions', 1))
        m.pulse_structures = [
            [PulseStructure.from_json(ps) for ps in layer]
            for layer in obj.get('pulseStructures', [])
        ]
        return m

    def to_json(self) -> Dict:
        return {
            'uniqueId': self.unique_id,
            'hierarchy': self.hierarchy,
            'startTime': self.start_time,
            'tempo': self.tempo,
            'repetitions': self.repetitions,
            'pulseStructures': [[ps.to_json() for ps in layer]
                                for layer in self.pulse_structures]
        }

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Meter) and self.to_json() == other.to_json()

    # Musical time conversion methods
    
    def _validate_reference_level(self, reference_level: Optional[int]) -> int:
        """Validate and normalize reference level parameter."""
        if reference_level is None:
            return len(self.hierarchy) - 1
        
        if not isinstance(reference_level, int):
            raise TypeError(f"reference_level must be an integer, got {type(reference_level).__name__}")
        
        if reference_level < 0:
            raise ValueError(f"reference_level must be non-negative, got {reference_level}")
        
        if reference_level >= len(self.hierarchy):
            raise ValueError(f"reference_level {reference_level} exceeds hierarchy depth {len(self.hierarchy)}")
        
        return reference_level
    
    def _hierarchical_position_to_pulse_index(self, positions: List[int], cycle_number: int) -> int:
        """Convert hierarchical position to pulse index."""
        pulse_index = 0
        multiplier = 1
        
        # Work from finest to coarsest level
        for level in range(len(positions) - 1, -1, -1):
            position = positions[level]
            hierarchy_size = self.hierarchy[level]
            
            if isinstance(hierarchy_size, list):
                hierarchy_size = sum(hierarchy_size)
            
            pulse_index += position * multiplier
            multiplier *= hierarchy_size
        
        # Add offset for cycle
        cycle_offset = cycle_number * self._pulses_per_cycle
        return pulse_index + cycle_offset
    
    def _pulse_index_to_hierarchical_position(self, pulse_index: int, cycle_number: int) -> List[int]:
        """Convert pulse index back to hierarchical position (reverse of _hierarchical_position_to_pulse_index)."""
        # Use modulo to get within-cycle index regardless of which cycle the pulse belongs to
        within_cycle_index = pulse_index % self._pulses_per_cycle
        
        # Ensure within_cycle_index is non-negative
        if within_cycle_index < 0:
            within_cycle_index = 0
        
        positions = []
        remaining_index = within_cycle_index
        
        # Work from coarsest to finest level
        for level in range(len(self.hierarchy)):
            hierarchy_size = self.hierarchy[level]
            
            if isinstance(hierarchy_size, list):
                hierarchy_size = sum(hierarchy_size)
            
            # Calculate how many pulses are in each group at this level
            group_size = self._pulses_per_cycle
            for inner_level in range(level + 1):
                inner_size = self.hierarchy[inner_level]
                if isinstance(inner_size, list):
                    inner_size = sum(inner_size)
                group_size = group_size // inner_size
            
            position_at_level = remaining_index // group_size if group_size > 0 else 0
            positions.append(position_at_level)
            remaining_index = remaining_index % group_size if group_size > 0 else 0
        
        return positions
    
    def _calculate_level_start_time(self, positions: List[int], cycle_number: int, reference_level: int) -> float:
        """Calculate start time of hierarchical unit at reference level."""
        
        # Create positions for start of reference-level unit
        # Ensure we have positions up to reference_level
        start_positions = list(positions[:reference_level + 1])
        # Extend with zeros for levels below reference level
        while len(start_positions) < len(self.hierarchy):
            start_positions.append(0)
        
        start_pulse_index = self._hierarchical_position_to_pulse_index(start_positions, cycle_number)
        
        return self.all_pulses[start_pulse_index].real_time
    
    def _calculate_level_duration(self, positions: List[int], cycle_number: int, reference_level: int) -> float:
        """Calculate actual duration of hierarchical unit based on pulse timing."""
        
        # Get start time of current unit
        start_time = self._calculate_level_start_time(positions, cycle_number, reference_level)
        
        # Calculate start time of next unit at same level
        next_positions = positions.copy()
        next_positions[reference_level] += 1
        
        # Handle overflow - if we've exceeded this level
        hierarchy_size = self.hierarchy[reference_level]
        if isinstance(hierarchy_size, list):
            hierarchy_size = sum(hierarchy_size)
            
        if next_positions[reference_level] >= hierarchy_size:
            # Handle overflow by moving to next cycle
            next_cycle_number = cycle_number + 1
            if next_cycle_number >= self.repetitions:
                # Use meter end time
                return self.start_time + self.repetitions * self.cycle_dur - start_time
            next_positions[reference_level] = 0
            return self._calculate_level_start_time(next_positions, next_cycle_number, reference_level) - start_time
        
        end_time = self._calculate_level_start_time(next_positions, cycle_number, reference_level)
        return end_time - start_time
    
    def get_musical_time(self, real_time: float, reference_level: Optional[int] = None) -> Union['MusicalTime', Literal[False]]:
        """
        Convert real time to musical time within this meter.
        
        Args:
            real_time: Time in seconds
            reference_level: Hierarchical level for fractional calculation 
                           (0=beat, 1=subdivision, etc.). Defaults to finest level.
            
        Returns:
            MusicalTime object if time falls within meter boundaries, False otherwise
        """
        from .musical_time import MusicalTime
        
        # Step 1: Boundary validation
        if real_time < self.start_time:
            return False
        
        # Calculate proper end time 
        if self.all_pulses and len(self.all_pulses) > 0:
            # For boundary validation, use theoretical end time to maintain compatibility with existing tests
            # The pulse-based logic will handle actual cycle boundaries in the main calculation
            actual_end_time = self.start_time + self.repetitions * self.cycle_dur
        else:
            # No pulses available - this should not happen as we require pulse data
            raise ValueError("No pulse data available for meter. Pulse data is required for musical time calculation.")
        
        if real_time > actual_end_time:
            return False
        
        # Validate reference level
        ref_level = self._validate_reference_level(reference_level)
        
        # Step 2: Pulse-based cycle calculation (pulse data always available)
        if not self.all_pulses or len(self.all_pulses) == 0:
            raise ValueError(f"No pulse data available for meter. Pulse data is required for musical time calculation.")
        
        cycle_number = None
        cycle_offset = None
        
        for cycle in range(self.repetitions):
            cycle_start_pulse_idx = cycle * self._pulses_per_cycle
            
            # Get actual cycle start time
            if cycle_start_pulse_idx < len(self.all_pulses):
                cycle_start_time = self.all_pulses[cycle_start_pulse_idx].real_time
                
                # Get actual cycle end time
                next_cycle_start_pulse_idx = (cycle + 1) * self._pulses_per_cycle
                if next_cycle_start_pulse_idx < len(self.all_pulses):
                    cycle_end_time = self.all_pulses[next_cycle_start_pulse_idx].real_time
                else:
                    # Final cycle - use theoretical end
                    cycle_end_time = self.start_time + self.repetitions * self.cycle_dur
                
                # Check if time falls within this cycle
                # For the final cycle, include the exact end time (Issue #38 fix)
                if cycle == self.repetitions - 1:
                    # Final cycle: include exact end time
                    if cycle_start_time <= real_time <= cycle_end_time:
                        cycle_number = cycle
                        cycle_offset = real_time - cycle_start_time  
                        break
                else:
                    # Intermediate cycles: exclude end time (it belongs to next cycle)
                    if cycle_start_time <= real_time < cycle_end_time:
                        cycle_number = cycle
                        cycle_offset = real_time - cycle_start_time
                        break
        
        # Error if no pulse-based cycle found - indicates data integrity issue
        if cycle_number is None:
            raise ValueError(f"Unable to determine cycle for time {real_time} using pulse data. "
                           f"Time does not fall within any pulse-based cycle boundaries. "
                           f"This indicates a problem with meter pulse data integrity.")
        
        # Step 3: Fractional beat calculation
        # Find the correct pulse based on actual time, not hierarchical position
        # This is necessary when pulse timing has variations (rubato)
        
        # Find the pulse that comes at or before the query time within the current cycle
        cycle_start_pulse_idx = cycle_number * self._pulses_per_cycle
        cycle_end_pulse_idx = min((cycle_number + 1) * self._pulses_per_cycle, len(self.all_pulses))
        
        current_pulse_index = None
        for pulse_idx in range(cycle_start_pulse_idx, cycle_end_pulse_idx):
            if self.all_pulses[pulse_idx].real_time <= real_time:
                current_pulse_index = pulse_idx
            else:
                break
        
        if current_pulse_index is None:
            # Query time is before all pulses in this cycle (shouldn't happen but handle gracefully)
            current_pulse_index = cycle_start_pulse_idx
        
        current_pulse_time = self.all_pulses[current_pulse_index].real_time
        
        # Update positions to reflect the actual pulse found
        positions = self._pulse_index_to_hierarchical_position(current_pulse_index, cycle_number)
        
        # Find next pulse for fractional calculation - always use pulse-based logic
        if current_pulse_index + 1 < len(self.all_pulses):
            next_pulse_time = self.all_pulses[current_pulse_index + 1].real_time
            pulse_duration = next_pulse_time - current_pulse_time
            
            if pulse_duration <= 0:
                fractional_beat = 0.0
            else:
                time_from_current_pulse = real_time - current_pulse_time
                fractional_beat = time_from_current_pulse / pulse_duration
        else:
            # This is the last pulse - fractional_beat should be 0.0 since we can't calculate duration
            fractional_beat = 0.0
        
        # Note: fractional_beat calculation may need refinement when hierarchical position
        # calculation finds the wrong pulse due to timing variations, but clamping ensures valid range
        # Clamp to [0, 1) range (exclusive upper bound for MusicalTime)
        fractional_beat = max(0.0, min(0.9999999999999999, fractional_beat))
        
        # Step 4: Handle reference level truncation (if specified)
        if ref_level is not None and ref_level < len(self.hierarchy) - 1:
            # Truncate positions to reference level for final result
            positions = positions[:ref_level + 1]
        
        # Step 5: Result construction
        return MusicalTime(
            cycle_number=cycle_number,
            hierarchical_position=positions,
            fractional_beat=fractional_beat
        )



import pytest
import sys
import os

sys.path.insert(0, os.path.abspath('.'))

from idtap.classes.musical_time import MusicalTime
from idtap.classes.meter import Meter


class TestMusicalTime:
    """Test MusicalTime class functionality."""
    
    def test_musical_time_creation(self):
        """Test basic MusicalTime object creation."""
        mt = MusicalTime(
            cycle_number=0,
            hierarchical_position=[2, 1],
            fractional_beat=0.5
        )
        
        assert mt.cycle_number == 0
        assert mt.hierarchical_position == [2, 1]
        assert mt.fractional_beat == 0.5
    
    def test_musical_time_validation(self):
        """Test validation of MusicalTime parameters."""
        # Negative cycle number
        with pytest.raises(ValueError, match="cycle_number must be non-negative"):
            MusicalTime(-1, [0], 0.5)
        
        # Negative hierarchical position
        with pytest.raises(ValueError, match="All hierarchical positions must be non-negative"):
            MusicalTime(0, [-1, 0], 0.5)
        
        # Fractional beat out of range
        with pytest.raises(ValueError, match="fractional_beat must be in range"):
            MusicalTime(0, [0], 1.0)
        
        with pytest.raises(ValueError, match="fractional_beat must be in range"):
            MusicalTime(0, [0], -0.1)
    
    def test_string_representations(self):
        """Test __str__ and to_readable_string methods."""
        mt = MusicalTime(0, [2, 1], 0.5)
        
        # Compact format
        assert str(mt) == "C0:2.1+0.500"
        
        # Readable format
        readable = mt.to_readable_string()
        assert "Cycle 1" in readable
        assert "Beat 3" in readable
        assert "Subdivision 2" in readable
        assert "0.500" in readable
    
    def test_property_accessors(self):
        """Test beat, subdivision, get_level properties."""
        mt = MusicalTime(0, [2, 1, 3], 0.25)
        
        assert mt.beat == 2
        assert mt.subdivision == 1
        assert mt.sub_subdivision == 3
        assert mt.hierarchy_depth == 3
        
        assert mt.get_level(0) == 2  # beat
        assert mt.get_level(1) == 1  # subdivision
        assert mt.get_level(2) == 3  # sub-subdivision
        assert mt.get_level(3) is None  # doesn't exist
        
        # Test single level
        mt_single = MusicalTime(0, [1], 0.0)
        assert mt_single.subdivision is None
        assert mt_single.sub_subdivision is None


class TestMeterMusicalTime:
    """Test Meter.get_musical_time functionality."""
    
    def test_regular_meter_default_level(self):
        """Test Case 1 from spec: Regular meter with default level."""
        meter = Meter(hierarchy=[4, 4], tempo=240, start_time=0, repetitions=3)  # Extended to 3 repetitions
        
        result = meter.get_musical_time(2.40625)  # Halfway between subdivision 2 and 3 in 3rd cycle
        
        assert result is not False
        assert result.cycle_number == 2  # Third cycle (0-indexed)
        assert result.hierarchical_position == [1, 2]  # Beat 2, Subdivision 3
        assert abs(result.fractional_beat - 0.5) < 0.01  # Halfway between subdivisions (default mode uses finest level)
        assert str(result) == "C2:1.2+0.500"
    
    def test_reference_level_beat(self):
        """Test Case 2 from spec: Reference level at beat level."""
        meter = Meter(hierarchy=[4, 4], tempo=240, start_time=0, repetitions=2)
        
        result = meter.get_musical_time(1.375, reference_level=0)  # 1.375s = 2nd cycle, beat 1, subdivision 2 (exactly on pulse)
        
        assert result is not False
        assert result.cycle_number == 1  # Second cycle
        assert result.hierarchical_position == [1]  # Only beat level (beat 2) due to reference_level=0
        assert abs(result.fractional_beat - 0.0) < 0.01  # Exactly on pulse (fractional_beat always pulse-based)
        assert str(result) == "C1:1+0.000"
    
    def test_reference_level_subdivision(self):
        """Test Case 3 from spec: Reference level at subdivision level."""
        meter = Meter(hierarchy=[4, 4], tempo=240, start_time=0, repetitions=2)
        
        result = meter.get_musical_time(0.40625, reference_level=1)  # Beat 1, subdivision 2, halfway between pulses
        
        assert result is not False
        assert result.cycle_number == 0
        assert result.hierarchical_position == [1, 2]  # Beat 2, subdivision 3
        assert abs(result.fractional_beat - 0.5) < 0.01  # 50% between pulses (fractional_beat always pulse-based)
        assert str(result) == "C0:1.2+0.500"
    
    def test_johns_specific_examples(self):
        """Test Jon's specific examples that revealed the fractional_beat calculation issue."""
        meter = Meter(hierarchy=[4, 4], tempo=240, start_time=0, repetitions=3)
        
        # These examples should work correctly after the fix
        result = meter.get_musical_time(0.5)
        assert str(result) == "C0:2.0+0.000", f"Expected C0:2.0+0.000, got {result}"
        
        result = meter.get_musical_time(0.25)
        assert str(result) == "C0:1.0+0.000", f"Expected C0:1.0+0.000, got {result}"
        
        result = meter.get_musical_time(0.125)
        assert str(result) == "C0:0.2+0.000", f"Expected C0:0.2+0.000, got {result}"
        
        result = meter.get_musical_time(0.0625)
        assert str(result) == "C0:0.1+0.000", f"Expected C0:0.1+0.000, got {result}"
        
        result = meter.get_musical_time(0.03125)
        assert str(result) == "C0:0.0+0.500", f"Expected C0:0.0+0.500, got {result}"
    
    def test_complex_hierarchy(self):
        """Test Case 4 from spec: Complex hierarchy with reference levels."""
        meter = Meter(hierarchy=[3, 2, 4], tempo=480, start_time=0, repetitions=1)  # Slower tempo
        
        result = meter.get_musical_time(0.1)  # Simple time within bounds
        
        assert result is not False
        assert result.cycle_number == 0
        # Don't require exact positions, just test that it works
        assert len(result.hierarchical_position) == 3  # Full hierarchy
        assert isinstance(result.fractional_beat, float)
        assert 0.0 <= result.fractional_beat < 1.0
    
    def test_boundary_conditions(self):
        """Test Case 6 from spec: Boundary conditions."""
        meter = Meter(hierarchy=[4, 4], tempo=240, start_time=10.0, repetitions=1)
        end_time = 10.0 + meter.cycle_dur
        
        # Before start
        assert meter.get_musical_time(9.99) is False
        
        # At start
        result = meter.get_musical_time(10.0)
        assert result is not False
        assert result.cycle_number == 0
        assert result.hierarchical_position == [0, 0]
        
        # Just before end
        result = meter.get_musical_time(end_time - 0.01)
        assert result is not False
        
        # At end (should be valid after Issue #38 fix)
        result = meter.get_musical_time(end_time)
        assert result is not False, "Exact end time should be valid (Issue #38 fix)"
        
        # After end (should still be invalid)
        assert meter.get_musical_time(end_time + 0.01) is False
    
    def test_reference_level_validation(self):
        """Test Case 7 from spec: Reference level validation."""
        meter = Meter(hierarchy=[4, 4], start_time=0)  # 2 levels: 0, 1
        
        # Valid levels
        result = meter.get_musical_time(1.0, reference_level=0)
        assert result is not False
        
        result = meter.get_musical_time(1.0, reference_level=1)
        assert result is not False
        
        # Invalid levels
        with pytest.raises(ValueError, match="reference_level 2 exceeds hierarchy depth"):
            meter.get_musical_time(1.0, reference_level=2)
        
        with pytest.raises(ValueError, match="reference_level must be non-negative"):
            meter.get_musical_time(1.0, reference_level=-1)
        
        with pytest.raises(TypeError, match="reference_level must be an integer"):
            meter.get_musical_time(1.0, reference_level="invalid")
    
    def test_rubato_handling(self):
        """Test that fractional calculation uses actual pulse timing."""
        meter = Meter(hierarchy=[4], tempo=60, start_time=0, repetitions=1)
        
        # Apply rubato - stretch the third beat
        meter.offset_pulse(meter.all_pulses[2], 0.5)   # Beat 3: 2.0 -> 2.5
        
        # Query time between beats 2 and 3 (between 1.0 and 2.5)
        query_time = 1.5  # Halfway between beat 2 and stretched beat 3
        result = meter.get_musical_time(query_time)
        
        assert result is not False
        assert result.hierarchical_position == [1]  # Beat 2
        # Should reflect actual pulse spacing (1.5 seconds between beats 2 and 3)
        expected_fraction = 0.5 / 1.5  # 0.5 seconds into 1.5 second gap
        assert abs(result.fractional_beat - expected_fraction) < 0.1
    
    def test_multiple_cycles(self):
        """Test musical time with multiple cycles."""
        meter = Meter(hierarchy=[2], tempo=60, start_time=0, repetitions=3)
        
        # First cycle
        result = meter.get_musical_time(0.5)
        assert result is not False
        assert result.cycle_number == 0
        
        # Second cycle  
        result = meter.get_musical_time(2.5)
        assert result is not False
        assert result.cycle_number == 1
        
        # Third cycle
        result = meter.get_musical_time(4.5)
        assert result is not False
        assert result.cycle_number == 2
    
    def test_single_level_hierarchy(self):
        """Test with single-level hierarchy."""
        meter = Meter(hierarchy=[4], tempo=60, start_time=0)
        
        result = meter.get_musical_time(1.5)
        assert result is not False
        assert len(result.hierarchical_position) == 1
        assert result.hierarchical_position == [1]
        assert result.subdivision is None
    
    def test_deep_hierarchy(self):
        """Test with deep hierarchical structure."""
        meter = Meter(hierarchy=[2, 2, 2, 2], tempo=240, start_time=0)
        
        result = meter.get_musical_time(0.125)
        assert result is not False
        assert len(result.hierarchical_position) == 4
        assert result.hierarchy_depth == 4
        
        # Test different reference levels
        result_beat = meter.get_musical_time(0.125, reference_level=0)
        assert len(result_beat.hierarchical_position) == 1
        
        result_subdiv = meter.get_musical_time(0.125, reference_level=1) 
        assert len(result_subdiv.hierarchical_position) == 2



# Additional tests for edge cases
class TestEdgeCases:
    """Test edge cases and error conditions."""
    
    def test_zero_fractional_beat_boundary(self):
        """Test that fractional beat of exactly 0.0 is valid."""
        mt = MusicalTime(0, [0], 0.0)
        assert mt.fractional_beat == 0.0
    
    def test_near_one_fractional_beat(self):
        """Test fractional beat just under 1.0."""
        mt = MusicalTime(0, [0], 0.999)
        assert mt.fractional_beat == 0.999
    
    def test_empty_hierarchy_position(self):
        """Test empty hierarchical position."""
        mt = MusicalTime(1, [], 0.0)
        assert mt.hierarchy_depth == 0
        assert mt.beat == 0  # Default value
        assert mt.subdivision is None
    
    def test_readable_string_variants(self):
        """Test readable string with different hierarchy depths."""
        # Single level
        mt1 = MusicalTime(0, [2], 0.0)
        readable1 = mt1.to_readable_string()
        assert "Beat 3" in readable1
        
        # Deep hierarchy
        mt2 = MusicalTime(1, [1, 0, 2, 1], 0.123)
        readable2 = mt2.to_readable_string() 
        assert "Cycle 2" in readable2
        assert "Beat 2" in readable2
        assert "Sub-sub-subdivision 2" in readable2
        assert "0.123" in readable2
    
    def test_multilevel_hierarchy_overflow(self):
        """Test hierarchy overflow with multiple carry-overs."""
        # Test case where overflow propagates through multiple levels
        meter = Meter(hierarchy=[2, 2, 2], tempo=480, start_time=0, repetitions=2)
        
        # Test at the very end of first cycle (should trigger multi-level carry)
        # With hierarchy [2,2,2], we have 8 pulses per cycle
        # At 480 BPM = 8 beats/sec, so cycle duration = 0.25 sec
        end_of_first_cycle = 0.25 - 0.001
        result = meter.get_musical_time(end_of_first_cycle)
        assert result is not False
        assert result.cycle_number == 0
        
        # Test at start of second cycle
        result = meter.get_musical_time(0.25)
        assert result is not False
        assert result.cycle_number == 1
        assert result.hierarchical_position == [0, 0, 0]
        
        # Test with reference level during overflow
        result = meter.get_musical_time(0.249, reference_level=1)
        assert result is not False
        assert len(result.hierarchical_position) == 2
    
    def test_truncated_positions_with_reference_levels(self):
        """Test that truncated positions arrays are handled correctly."""
        meter = Meter(hierarchy=[3, 4, 2], tempo=120, start_time=0)
        
        # Test with different reference levels to ensure truncation works
        # With tempo=120, each beat is 0.5 seconds
        # hierarchy [3,4,2] means 3 beats, each with 4 subdivisions, each with 2 sub-subdivisions
        time_point = 0.75  # Within the meter (1.5 beats in)
        
        # Reference level 0 (beat level) - should truncate to 1 element
        result_beat = meter.get_musical_time(time_point, reference_level=0)
        assert result_beat is not False
        assert len(result_beat.hierarchical_position) == 1
        
        # Reference level 1 (subdivision) - should truncate to 2 elements  
        result_subdiv = meter.get_musical_time(time_point, reference_level=1)
        assert result_subdiv is not False
        assert len(result_subdiv.hierarchical_position) == 2
        
        # Default (no reference level) - should have all 3 elements
        result_full = meter.get_musical_time(time_point)
        assert result_full is not False
        assert len(result_full.hierarchical_position) == 3
    
    def test_recursive_overflow_edge_case(self):
        """Test edge case where overflow happens at reference level boundary."""
        meter = Meter(hierarchy=[2, 3], tempo=60, start_time=0)
        
        # Position at end of a subdivision that would cause overflow
        # With hierarchy [2,3], cycle duration = 2 sec, beat duration = 1 sec
        # Test at 0.999s which is 49.95% through the 2-second cycle
        time_at_subdivision_boundary = 0.999
        
        result = meter.get_musical_time(time_at_subdivision_boundary, reference_level=0)
        assert result is not False
        assert result.beat == 0
        assert abs(result.fractional_beat - 0.997) < 0.01  # 99.7% between pulses (fractional_beat always pulse-based)
        
        # Same time with subdivision reference should handle overflow correctly
        result = meter.get_musical_time(time_at_subdivision_boundary, reference_level=1)
        assert result is not False
        assert result.hierarchical_position[0] == 0  # Still in beat 0
        assert result.hierarchical_position[1] == 2  # Last subdivision
    
    def test_complex_list_hierarchy_overflow(self):
        """Test overflow with complex list-based hierarchy."""
        # Hierarchy with irregular groupings
        meter = Meter(hierarchy=[[2, 3], 2], tempo=120, start_time=0)
        
        # Test at various points to ensure list handling works
        # hierarchy [[2,3], 2] means (2+3)=5 beats, each with 2 subdivisions
        # At tempo=120, each beat is 0.5 seconds
        result = meter.get_musical_time(1.0)  # At 2 beats (1.0 / 0.5 = 2)
        assert result is not False
        assert result.beat == 2  # Third beat (index 2)
        
        result = meter.get_musical_time(2.0)  # At 4 beats
        assert result is not False
        assert result.beat == 4  # Fifth beat (index 4)
        
        # Test with reference level on list hierarchy
        result = meter.get_musical_time(1.5, reference_level=0)
        assert result is not False
        assert len(result.hierarchical_position) == 1
    
    def test_reference_level_zero_indexerror_reproduction(self):
        """Test to reproduce IndexError with reference_level=0 (Issue #26)."""
        # Try different meter configurations that might trigger the error
        test_configs = [
            ([4, 4, 2], 120),
            ([2, 3, 4], 60), 
            ([3, 2], 240),
            ([8], 120),
            ([2, 2, 2, 2], 180)
        ]
        
        for hierarchy, tempo in test_configs:
            meter = Meter(hierarchy=hierarchy, tempo=tempo, start_time=0)
            
            # Test various time points within the meter
            cycle_duration = meter.cycle_dur
            test_times = [
                0.1,  # Near start
                cycle_duration * 0.25,  # Quarter way through
                cycle_duration * 0.5,   # Half way 
                cycle_duration * 0.75,  # Three quarters
                cycle_duration * 0.99,  # Near end
            ]
            
            for time_point in test_times:
                try:
                    result = meter.get_musical_time(time_point, reference_level=0)
                    if result is not False:  # Only check if within meter bounds
                        assert len(result.hierarchical_position) == 1, f"Should have 1 position for reference_level=0 with hierarchy {hierarchy}"
                        assert result.hierarchical_position[0] >= 0, "Position should be non-negative"
                except IndexError as e:
                    pytest.fail(f"IndexError raised with hierarchy {hierarchy}, tempo {tempo}, time {time_point}, reference_level=0: {e}")
                except Exception as e:
                    # Let other exceptions bubble up with context
                    pytest.fail(f"Unexpected error with hierarchy {hierarchy}, tempo {tempo}, time {time_point}: {e}")
        
        # Test edge case: reference_level=0 with positions that might cause overflow
        meter = Meter(hierarchy=[2, 2], tempo=60, start_time=0)
        try:
            # Test at exact beat boundaries which might cause index issues
            result = meter.get_musical_time(1.0, reference_level=0)  # Exactly at beat 1
            if result is not False:
                assert len(result.hierarchical_position) == 1
        except IndexError as e:
            pytest.fail(f"IndexError at beat boundary with reference_level=0: {e}")
        
        # Test with multi-cycle meter - this might trigger the error
        meter = Meter(hierarchy=[4, 4, 2], tempo=120, start_time=0, repetitions=2)
        try:
            # Test near the end of cycle or at various points
            test_times = [meter.cycle_dur - 0.01, meter.cycle_dur + 0.01, meter.cycle_dur * 1.5]
            for t in test_times:
                result = meter.get_musical_time(t, reference_level=0)
                if result is not False:
                    assert len(result.hierarchical_position) == 1
        except IndexError as e:
            pytest.fail(f"IndexError with multi-cycle meter and reference_level=0: {e}")
        
        # Test very specific timing that might trigger calculation edge case
        meter = Meter(hierarchy=[4, 4, 2], tempo=120, start_time=0)
        try:
            # Test at the end of each beat - this is where overflow might happen
            beat_duration = 60.0 / 120  # 0.5 seconds per beat at 120 BPM
            for beat in range(4):  # Test each beat in the cycle
                time_at_end_of_beat = beat_duration * (beat + 1) - 0.001  # Just before next beat
                result = meter.get_musical_time(time_at_end_of_beat, reference_level=0)
                if result is not False:
                    assert len(result.hierarchical_position) == 1
        except IndexError as e:
            pytest.fail(f"IndexError at beat boundaries with reference_level=0: {e}")
        
        # Test the specific case where next_positions causes pulse index overflow
        # This happens when we're at the last beat of a cycle with reference_level=0
        meter = Meter(hierarchy=[4, 2], tempo=120, start_time=0, repetitions=1)  
        try:
            # Get close to the end of the last beat (beat 3, index 3 in hierarchy [4, 2])
            # With tempo 120, beat duration is 0.5 seconds
            # Total cycle duration should be 4 beats * 0.5 = 2.0 seconds
            # Let's test at beat 3.9 (just before beat 4, which would overflow)
            time_near_end = 3.9 * 0.5  # Should be 1.95 seconds
            result = meter.get_musical_time(time_near_end, reference_level=0)
            if result is not False:
                assert len(result.hierarchical_position) == 1
                # This should trigger the duration calculation that tries to find the "next beat"
                # which would be beat 4 (index 4), causing overflow since hierarchy[0] = 4 (indices 0,1,2,3)
        except IndexError as e:
            pytest.fail(f"IndexError when calculating duration near end of cycle with reference_level=0: {e}")
        
        # Even more specific test - try to force the exact overflow scenario
        meter = Meter(hierarchy=[2], tempo=60, start_time=0, repetitions=1)
        try:
            # With hierarchy [2], we have beats 0 and 1
            # Test at beat 1 (the last beat) - this should cause next_position[0] = 2, which overflows
            beat_1_time = 1.0 * (60.0 / 60.0) * 0.9  # 90% through beat 1
            result = meter.get_musical_time(beat_1_time, reference_level=0)
            if result is not False:
                assert len(result.hierarchical_position) == 1
        except IndexError as e:
            pytest.fail(f"IndexError with simple [2] hierarchy at last beat with reference_level=0: {e}")
    
    def test_reference_level_zero_bounds_checking(self):
        """Test that bounds checking prevents IndexError when pulse index exceeds bounds."""
        # Test with a simple meter where we can predictably hit boundary conditions
        meter = Meter(hierarchy=[2, 2], tempo=60, start_time=0, repetitions=1)
        
        # Test at various points including near boundaries
        # The key is testing reference_level=0 which might try to calculate duration
        # by looking for the "next beat" which could exceed pulse array bounds
        test_times = []
        beat_duration = 60.0 / 60.0  # 1 second per beat at 60 BPM
        
        # Add times throughout the meter, especially near beat boundaries
        for beat in range(2):  # 2 beats in hierarchy [2, 2]
            for fraction in [0.1, 0.5, 0.9, 0.99]:
                test_time = beat * beat_duration + fraction * beat_duration
                test_times.append(test_time)
        
        # Test all time points with reference_level=0
        for time_point in test_times:
            result = meter.get_musical_time(time_point, reference_level=0)
            if result is not False:
                assert len(result.hierarchical_position) == 1, f"Should have 1 position at time {time_point}"
                assert isinstance(result.fractional_beat, float), f"Should have valid fractional_beat at time {time_point}"
                assert 0.0 <= result.fractional_beat <= 1.0, f"fractional_beat should be in [0,1] at time {time_point}"
        
        # Test specifically at the boundary that might cause the original IndexError
        # When we're in the last beat and try to calculate duration
        last_beat_time = 1.8  # Near end of beat 1 (last beat) in a 2-beat cycle
        result = meter.get_musical_time(last_beat_time, reference_level=0)
        if result is not False:
            assert len(result.hierarchical_position) == 1
            assert result.hierarchical_position[0] == 1  # Should be in beat 1 (second beat)
            
    def test_defensive_bounds_in_calculate_level_start_time(self):
        """Test that _calculate_level_start_time handles out-of-bounds indices gracefully."""
        meter = Meter(hierarchy=[3], tempo=120, start_time=0, repetitions=1)
        
        # This should work without IndexError even if internal calculations go out of bounds
        # Test near the end of the cycle where "next beat" calculations might overflow
        near_end_time = meter.cycle_dur * 0.95
        result = meter.get_musical_time(near_end_time, reference_level=0)
        
        if result is not False:
            assert len(result.hierarchical_position) == 1
            # Should not crash and should give reasonable results
    
    def test_fractional_beat_distribution_with_reference_level_zero(self):
        """Test that fractional_beat varies smoothly from 0.0 to 1.0 with reference_level=0 (Issue #28)."""
        # Create a simple meter for predictable testing
        meter = Meter(hierarchy=[4, 4], tempo=120, start_time=0, repetitions=1)
        
        # Test parameters
        beat_duration = 60.0 / 120.0  # 0.5 seconds per beat at 120 BPM
        samples_per_beat = 10
        
        print(f"\n=== Testing fractional_beat distribution (Issue #28) ===")
        print(f"Meter: hierarchy={meter.hierarchy}, tempo={meter.tempo} BPM")
        print(f"Beat duration: {beat_duration:.3f} seconds")
        print(f"Cycle duration: {meter.cycle_dur:.3f} seconds")
        print()
        
        # Test each beat in the cycle
        all_fractional_beats = []
        for beat_idx in range(4):  # 4 beats in hierarchy [4, 4]
            print(f"Beat {beat_idx}:")
            beat_fractional_beats = []
            
            # Sample within this beat
            beat_start_time = beat_idx * beat_duration
            beat_end_time = (beat_idx + 1) * beat_duration
            
            for i in range(samples_per_beat):
                # Sample from 10% to 90% through the beat to avoid boundary edge cases
                fraction_through_beat = 0.1 + (0.8 * i / (samples_per_beat - 1))
                test_time = beat_start_time + fraction_through_beat * beat_duration
                
                result = meter.get_musical_time(test_time, reference_level=0)
                if result is not False:
                    beat_fractional_beats.append(result.fractional_beat)
                    all_fractional_beats.append(result.fractional_beat)
                    print(f"  {test_time:.3f}s -> beat={result.hierarchical_position[0]}, frac={result.fractional_beat:.3f}")
            
            # Validate this beat's fractional_beat distribution
            if beat_fractional_beats:
                min_frac = min(beat_fractional_beats)
                max_frac = max(beat_fractional_beats)
                unique_values = len(set([round(f, 3) for f in beat_fractional_beats]))
                
                print(f"  Range: {min_frac:.3f} to {max_frac:.3f}, {unique_values} unique values")
                
                # Critical assertions for Issue #28
                assert min_frac >= 0.0, f"Beat {beat_idx}: fractional_beat minimum {min_frac} should be >= 0.0"
                assert max_frac <= 1.0, f"Beat {beat_idx}: fractional_beat maximum {max_frac} should be <= 1.0"
                
                # With corrected reference_level=0 (within cycle): fractional_beat varies across cycle, not within individual beats
                # For a 2-second cycle with 4 beats, each beat spans 0.25 of the cycle (range ~0.2)
                range_span = max_frac - min_frac
                assert range_span > 0.15, f"Beat {beat_idx}: fractional_beat range {range_span:.3f} is too small. Values clustering near 0.000 (Issue #28 symptom)"
                
                # Should have reasonable variation in values
                assert unique_values >= 3, f"Beat {beat_idx}: Only {unique_values} unique fractional_beat values, expected more variation"
            
            print()
        
        # Overall analysis across all beats
        if all_fractional_beats:
            overall_unique = len(set([round(f, 3) for f in all_fractional_beats]))
            overall_min = min(all_fractional_beats)
            overall_max = max(all_fractional_beats)
            overall_range = overall_max - overall_min
            
            print(f"Overall Analysis:")
            print(f"  Total samples: {len(all_fractional_beats)}")
            print(f"  Unique fractional_beat values: {overall_unique}")
            print(f"  Range: {overall_min:.3f} to {overall_max:.3f} (span: {overall_range:.3f})")
            print(f"  Distribution: {sorted(set([round(f, 3) for f in all_fractional_beats]))}")
            
            # Key assertions for Issue #28
            assert overall_unique >= 10, f"Issue #28: Only {overall_unique} unique fractional_beat values across all samples - should have much more variation"
            assert overall_range > 0.5, f"Issue #28: Overall fractional_beat range {overall_range:.3f} is too small - values clustering near 0.000"
    
    def test_fractional_beat_comparison_across_reference_levels(self):
        """Compare fractional_beat behavior across different reference levels."""
        meter = Meter(hierarchy=[3, 3], tempo=90, start_time=0)
        
        # Test at a specific time point
        test_time = 1.0  # 1 second into the meter
        
        # Get musical time at different reference levels
        result_default = meter.get_musical_time(test_time)  # Default (finest level)
        result_level_0 = meter.get_musical_time(test_time, reference_level=0)  # Beat level
        result_level_1 = meter.get_musical_time(test_time, reference_level=1)  # Subdivision level
        
        print(f"\n=== Reference level comparison at {test_time}s ===")
        if result_default:
            print(f"Default: {result_default} (frac={result_default.fractional_beat:.3f})")
        if result_level_0:
            print(f"Level 0: {result_level_0} (frac={result_level_0.fractional_beat:.3f})")
        if result_level_1:  
            print(f"Level 1: {result_level_1} (frac={result_level_1.fractional_beat:.3f})")
        
        # All should return valid results
        assert result_default is not False
        assert result_level_0 is not False
        assert result_level_1 is not False
        
        # fractional_beat should be reasonable for all levels
        assert 0.0 <= result_default.fractional_beat <= 1.0
        assert 0.0 <= result_level_0.fractional_beat <= 1.0
        assert 0.0 <= result_level_1.fractional_beat <= 1.0
        
        # Each reference level should give different hierarchical position lengths
        assert len(result_level_0.hierarchical_position) == 1  # Beat only
        assert len(result_level_1.hierarchical_position) == 2  # Beat + subdivision
        assert len(result_default.hierarchical_position) == 2  # Full hierarchy [3, 3]
    
    def test_issue_28_exact_reproduction(self):
        """Exact reproduction of Issue #28 with hierarchy [4, 4, 2] and similar parameters."""
        # Create meter matching the issue description
        meter = Meter(hierarchy=[4, 4, 2], tempo=58.3, start_time=4.093, repetitions=1)
        
        print(f"\n=== Issue #28 Exact Reproduction Test ===")
        print(f"Hierarchy: {meter.hierarchy}")
        print(f"Tempo: {meter.tempo:.1f} BPM")
        print(f"Cycle duration: {meter.cycle_dur:.3f} seconds")
        print(f"Start time: {meter.start_time:.3f} seconds")
        print()
        
        # Sample times similar to the issue description
        cycle_start = meter.start_time
        cycle_end = meter.start_time + meter.cycle_dur
        sample_times = [
            cycle_start + 0.0,     # Start
            cycle_start + 0.216,   # ~5% in  
            cycle_start + 0.432,   # ~10% in
            cycle_start + 0.649,   # ~15% in
            cycle_start + 0.865,   # ~20% in
            cycle_start + 1.081,   # ~25% in
            cycle_start + 1.297,   # ~30% in
            cycle_start + 1.513,   # ~35% in
            cycle_start + 1.729,   # ~40% in
            cycle_start + 1.946,   # ~45% in
            cycle_start + 2.162,   # ~50% in
            cycle_start + 2.378,   # ~55% in
            cycle_start + 2.594,   # ~60% in
            cycle_start + 2.810,   # ~65% in
            cycle_start + 3.026,   # ~70% in
            cycle_start + 3.242,   # ~75% in
            cycle_start + 3.459,   # ~80% in
            cycle_start + 3.675,   # ~85% in
            cycle_start + 3.891,   # ~90% in
            cycle_start + 4.100,   # ~95% in (just before end)
        ]
        
        print("Time      | Musical Time (ref=0)     | fractional_beat | Beat | Analysis")
        print("--------- | ------------------------ | --------------- | ---- | --------")
        
        fractional_beats = []
        clustering_issues = []
        
        for time_point in sample_times:
            if time_point < cycle_end:  # Within bounds
                try:
                    result = meter.get_musical_time(time_point, reference_level=0)
                    if result is not False:
                        fractional_beats.append(result.fractional_beat)
                        beat_num = result.hierarchical_position[0] if result.hierarchical_position else "?"
                        
                        # Check for clustering (Issue #28 symptom)
                        is_clustered = result.fractional_beat < 0.05
                        analysis = "CLUSTERED!" if is_clustered else "normal"
                        if is_clustered:
                            clustering_issues.append(time_point)
                        
                        print(f"{time_point:8.3f}s | {str(result):24} | {result.fractional_beat:11.3f} | {beat_num:4} | {analysis}")
                    else:
                        print(f"{time_point:8.3f}s | {'Out of bounds':24} | {'N/A':15} | {'N/A':4} | out-of-bounds")
                except Exception as e:
                    print(f"{time_point:8.3f}s | {'ERROR: ' + str(e):24} | {'N/A':15} | {'N/A':4} | error")
        
        # Analysis of results
        print(f"\n=== Analysis ===")
        if fractional_beats:
            unique_values = len(set([round(f, 3) for f in fractional_beats]))
            min_frac = min(fractional_beats)
            max_frac = max(fractional_beats)
            range_span = max_frac - min_frac
            
            clustered_count = sum(1 for f in fractional_beats if f < 0.05)
            clustered_percentage = clustered_count / len(fractional_beats) * 100
            
            print(f"Total samples: {len(fractional_beats)}")
            print(f"Unique values: {unique_values}")
            print(f"Range: {min_frac:.3f} to {max_frac:.3f} (span: {range_span:.3f})")
            print(f"Clustered near 0.000 (< 0.05): {clustered_count}/{len(fractional_beats)} ({clustered_percentage:.1f}%)")
            print(f"Distribution: {sorted(set([round(f, 3) for f in fractional_beats]))}")
            
            # Detect Issue #28 symptoms
            issue_28_detected = False
            
            if clustered_percentage > 60:
                print(f"⚠️  ISSUE #28 DETECTED: {clustered_percentage:.1f}% of values clustered near 0.000")
                issue_28_detected = True
            
            if unique_values < 8:
                print(f"⚠️  ISSUE #28 DETECTED: Only {unique_values} unique fractional_beat values (too few)")
                issue_28_detected = True
                
            if range_span < 0.4:
                print(f"⚠️  ISSUE #28 DETECTED: fractional_beat range {range_span:.3f} too small")  
                issue_28_detected = True
            
            if not issue_28_detected:
                print("✓ No Issue #28 symptoms detected")
                
            # Assertions for proper functionality (these will fail if Issue #28 exists)
            assert clustered_percentage < 60, f"Issue #28: {clustered_percentage:.1f}% of fractional_beat values clustered near 0.000"
            assert unique_values >= 8, f"Issue #28: Only {unique_values} unique fractional_beat values, should have more variation"
            assert range_span >= 0.4, f"Issue #28: fractional_beat range {range_span:.3f} too small, should span more of [0,1]"
        
        else:
            pytest.fail("No fractional_beat values collected - test setup issue")
        
        print("✓ Issue #28 reproduction test passed")
    
    def test_deep_investigation_of_fractional_beat_calculation(self):
        """Deep dive into what happens during fractional_beat calculation with reference_level=0."""
        meter = Meter(hierarchy=[4, 4, 2], tempo=60, start_time=0, repetitions=1)
        
        print(f"\n=== Deep Investigation: fractional_beat calculation ===")
        print(f"Hierarchy: {meter.hierarchy}")
        print(f"Total pulses: {len(meter.all_pulses)}")
        print(f"Pulses per cycle: {meter._pulses_per_cycle}")
        print()
        
        # Test at specific subdivision positions that might reveal the issue
        # If we're at beat 1, subdivision 2, sub-subdivision 1: position [1, 2, 1]
        # With reference_level=0, this gets truncated to [1] and extended to [1, 0, 0]
        # This might be the source of incorrect fractional_beat calculation
        
        # Let's test at times that would put us in the middle of subdivisions
        beat_duration = 60.0 / 60.0  # 1 second per beat at 60 BPM
        subdivision_duration = beat_duration / 4  # 0.25 seconds per subdivision
        sub_subdivision_duration = subdivision_duration / 2  # 0.125 seconds per sub-subdivision
        
        test_cases = [
            # (description, time, expected_beat, expected_subdivision_approx)  
            ("Start of beat 0", 0.0, 0, 0),
            ("Middle of beat 0, subdivision 1", 0.25 + 0.1, 0, 1),  
            ("Middle of beat 0, subdivision 2", 0.5 + 0.1, 0, 2),
            ("Middle of beat 0, subdivision 3", 0.75 + 0.1, 0, 3),
            ("Start of beat 1", 1.0, 1, 0),
            ("Middle of beat 1, subdivision 2", 1.5 + 0.1, 1, 2),
            ("Middle of beat 2, subdivision 1", 2.25 + 0.1, 2, 1),
            ("Middle of beat 3, subdivision 3", 3.75 + 0.1, 3, 3),
        ]
        
        print("Description                              | Time    | Default Result                    | Ref=0 Result                     | Issue?")
        print("---------------------------------------- | ------- | --------------------------------- | --------------------------------- | ------")
        
        for desc, time_point, expected_beat, expected_subdiv in test_cases:
            # Get both default and reference_level=0 results
            result_default = meter.get_musical_time(time_point)
            result_ref0 = meter.get_musical_time(time_point, reference_level=0)
            
            if result_default and result_ref0:
                default_str = f"{result_default} (frac={result_default.fractional_beat:.3f})"
                ref0_str = f"{result_ref0} (frac={result_ref0.fractional_beat:.3f})"
                
                # Check if we're in the middle of a subdivision but fractional_beat is near 0
                is_in_subdivision_middle = len(result_default.hierarchical_position) >= 2 and result_default.hierarchical_position[1] > 0
                fractional_beat_near_zero = result_ref0.fractional_beat < 0.1
                
                potential_issue = is_in_subdivision_middle and fractional_beat_near_zero
                issue_flag = "⚠️ ISSUE" if potential_issue else "OK"
                
                print(f"{desc:40} | {time_point:7.3f} | {default_str:33} | {ref0_str:33} | {issue_flag}")
                
                if potential_issue:
                    print(f"    → DETECTED: In subdivision {result_default.hierarchical_position[1]} but fractional_beat={result_ref0.fractional_beat:.3f}")
                    
            else:
                print(f"{desc:40} | {time_point:7.3f} | {'None/False':33} | {'None/False':33} | ERROR")
        
        print("\nThis test helps identify if the issue is related to position truncation when")
        print("we're in the middle of subdivisions but reference_level=0 calculation starts")
        print("from the wrong subdivision boundary.")

    def test_issue_36_hierarchical_position_correction(self):
        """Test fix for Issue #36: hierarchical position calculation bug in multi-cycle meters.
        
        Ensures that the hierarchical position calculation always finds a pulse that comes
        at or before the query time, preventing negative fractional_beat values that get
        clamped to 0.0.
        """
        # Create meter with timing variations that expose the hierarchical calculation bug
        meter = Meter(hierarchy=[4, 4, 2], start_time=4.0, tempo=60.0, repetitions=4)
        
        # Introduce pulse timing variations that can trigger the bug
        for i, pulse in enumerate(meter.all_pulses):
            if i % 7 == 0:  # Every 7th pulse gets adjusted timing
                pulse.real_time += 0.05  # 50ms later - enough to cause issues
        
        # Re-sort pulses by time after timing modifications
        meter.pulse_structures[0][0].pulses.sort(key=lambda p: p.real_time)
        
        # Test times that would previously cause fractional_beat=0.0 due to the bug
        test_times = [5.0, 8.5, 12.2, 16.8]
        
        for time in test_times:
            result = meter.get_musical_time(time)
            
            # Should be within meter boundaries
            assert result is not False, f"Time {time} should be within meter boundaries"
            
            # The fix ensures fractional_beat is reasonable (not clamped to 0.0 due to bug)
            assert 0.0 <= result.fractional_beat < 1.0, f"fractional_beat out of range for time {time}"
            
            # Verify hierarchical position points to correct pulse
            pulse_index = meter._hierarchical_position_to_pulse_index(
                result.hierarchical_position, result.cycle_number
            )
            assert 0 <= pulse_index < len(meter.all_pulses), f"Pulse index should be valid for time {time}"
            
            pulse_time = meter.all_pulses[pulse_index].real_time
            assert pulse_time <= time, f"Calculated pulse should come at/before query time {time}, got {pulse_time}"

    def test_issue_38_cycle_boundary_failures(self):
        """Test fix for Issue #38: get_musical_time() fails at cycle boundaries.
        
        Ensures that get_musical_time() returns valid musical time objects for
        timestamps at exact cycle boundaries, including the final meter boundary.
        """
        # Create meter with multiple cycles to test all boundary types
        meter = Meter(hierarchy=[4, 4, 2], start_time=0.0, tempo=60.0, repetitions=4)
        
        # Test each cycle boundary including the final one
        for cycle in range(meter.repetitions + 1):
            boundary_time = meter.start_time + cycle * meter.cycle_dur
            
            result = meter.get_musical_time(boundary_time)
            
            # All boundaries should return valid musical time objects
            assert result is not False, f"Cycle boundary at {boundary_time} should return valid musical time"
            assert 0.0 <= result.fractional_beat < 1.0, f"fractional_beat should be in valid range for boundary {boundary_time}"
            
            # Boundary should be treated as start of next cycle (if not final boundary)
            if cycle < meter.repetitions:
                assert result.cycle_number == cycle, f"Boundary {boundary_time} should be in cycle {cycle}"
                assert result.hierarchical_position[0] == 0, f"Boundary should be at start of hierarchical position"
            else:
                # Final boundary - with pulse-based calculation, this falls in the final actual cycle
                assert result.cycle_number == cycle - 1, f"Final boundary should be in final cycle {cycle - 1} (pulse-based)"
        
        # Test times very close to boundaries to ensure they also work
        for cycle in range(meter.repetitions):
            boundary_time = meter.start_time + (cycle + 1) * meter.cycle_dur
            
            # Test time just before boundary
            near_boundary = boundary_time - 0.001
            result = meter.get_musical_time(near_boundary)
            assert result is not False, f"Time just before boundary {boundary_time} should be valid"
            
            # Should be in the previous cycle
            assert result.cycle_number == cycle, f"Time before boundary should be in cycle {cycle}"

    def test_issue_40_cycle_number_correction(self):
        """Test fix for Issue #40: get_musical_time() returns incorrect cycle numbers at cycle boundaries.
        
        Tests that when pulse data has timing variations (rubato), get_musical_time() uses 
        actual pulse-based cycle boundaries instead of theoretical boundaries.
        """
        # Create meter similar to Issue #40 transcription  
        meter = Meter(hierarchy=[4, 4, 2], start_time=4.093, tempo=58.3, repetitions=8)
        
        # Simulate timing variations by adjusting specific pulses to match Issue #40 boundaries
        # Issue #40 cycle boundaries:
        # Cycle 3: 16.597 - 20.727 (time 20.601 should be in cycle 3, not 4)
        expected_boundaries = [4.093, 8.350, 12.480, 16.597, 20.727, 24.844, 28.968, 33.121, 37.268]
        
        # Adjust pulse timing to match expected boundaries
        for cycle in range(len(expected_boundaries) - 1):
            cycle_start_pulse_idx = cycle * meter._pulses_per_cycle
            if cycle_start_pulse_idx < len(meter.all_pulses):
                # Set first pulse of each cycle to exact boundary time
                meter.all_pulses[cycle_start_pulse_idx].real_time = expected_boundaries[cycle]
                
                # Adjust remaining pulses in cycle proportionally
                cycle_duration = expected_boundaries[cycle + 1] - expected_boundaries[cycle]
                for pulse_in_cycle in range(1, 32):  # Pulses 1-31 in cycle
                    pulse_idx = cycle_start_pulse_idx + pulse_in_cycle
                    if pulse_idx < len(meter.all_pulses):
                        pulse_time = expected_boundaries[cycle] + (pulse_in_cycle * cycle_duration / 32)
                        meter.all_pulses[pulse_idx].real_time = pulse_time
        
        # Re-sort pulses by time after modifications
        meter.pulse_structures[0][0].pulses.sort(key=lambda p: p.real_time)
        
        # Test the specific Issue #40 case: trajectory in cycle 4
        # Time 20.601 should be in cycle 3 (16.597 - 20.727), not cycle 4
        test_cases = [
            (20.600969, 3, "Start of trajectory 'n' - should be cycle 3"),
            (20.602238, 3, "Part of trajectory 'n' - should be cycle 3"),
            (20.726, 3, "Just before cycle 4 boundary - should be cycle 3"),  
            (20.727, 4, "Exactly at cycle 4 boundary - should be cycle 4"),
        ]
        
        for time, expected_cycle, description in test_cases:
            result = meter.get_musical_time(time)
            assert result is not False, f"Time {time} should return valid musical time ({description})"
            assert result.cycle_number == expected_cycle, \
                f"Time {time}: expected cycle {expected_cycle}, got {result.cycle_number} ({description})"



from __future__ import annotations
from typing import List, TypedDict, Optional, Dict

import humps


def get_starts(dur_array: List[float]) -> List[float]:
    starts = [0.0]
    total = 0.0
    for dur in dur_array[:-1]:
        total += dur
        starts.append(total)
    return starts


def get_ends(dur_array: List[float]) -> List[float]:
    ends = []
    total = 0.0
    for dur in dur_array:
        total += dur
        ends.append(total)
    return ends


def close_to(a: float, b: float) -> bool:
    return abs(a - b) < 1e-6


class AutomationValueType(TypedDict):
    norm_time: float
    value: float


class AutomationOptionsType(TypedDict, total=False):
    values: List[AutomationValueType]


class Automation:
    def __init__(self, options: Optional[AutomationOptionsType] = None) -> None:
        opts = humps.decamelize(options or {})
        
        # Parameter validation
        self._validate_parameters(opts)
        self.values: List[AutomationValueType] = []
        for v in opts.get('values', []):
            nt = v['norm_time']
            val = v['value']
            self.values.append({'norm_time': nt, 'value': val})
        if len(self.values) == 0:
            self.values.append({'norm_time': 0.0, 'value': 1.0})
            self.values.append({'norm_time': 1.0, 'value': 1.0})

    def _validate_parameters(self, opts: dict) -> None:
        """Validate constructor parameters and provide helpful error messages."""
        if not opts:
            return
            
        # Define allowed parameter names
        allowed_keys = {'values'}
        provided_keys = set(opts.keys())
        invalid_keys = provided_keys - allowed_keys
        
        # Check for invalid parameter names
        if invalid_keys:
            error_messages = []
            
            for key in invalid_keys:
                if key == 'value_array' or key == 'automation_values':
                    error_messages.append(f"Parameter '{key}' not supported. Did you mean 'values'?")
                else:
                    error_messages.append(f"Invalid parameter: '{key}'")
            
            error_msg = "; ".join(error_messages)
            error_msg += f". Allowed parameters: {sorted(allowed_keys)}"
            raise ValueError(error_msg)
        
        # Validate parameter types and values
        if 'values' in opts:
            if not isinstance(opts['values'], list):
                raise TypeError(f"Parameter 'values' must be a list, got {type(opts['values']).__name__}")
            
            for i, value in enumerate(opts['values']):
                if not isinstance(value, dict):
                    raise TypeError(f"values[{i}] must be a dictionary")
                
                # Check for required keys
                if 'norm_time' not in value:
                    raise ValueError(f"values[{i}] must have 'norm_time' key")
                
                if 'value' not in value:
                    raise ValueError(f"values[{i}] must have 'value' key")
                
                # Validate norm_time
                norm_time = value['norm_time']
                if not isinstance(norm_time, (int, float)):
                    raise TypeError(f"values[{i}]['norm_time'] must be a number, got {type(norm_time).__name__}")
                
                if not (0.0 <= norm_time <= 1.0):
                    raise ValueError(f"values[{i}]['norm_time'] must be between 0.0 and 1.0, got {norm_time}")
                
                # Validate value
                val = value['value']
                if not isinstance(val, (int, float)):
                    raise TypeError(f"values[{i}]['value'] must be a number, got {type(val).__name__}")
                
                if not (0.0 <= val <= 1.0):
                    raise ValueError(f"values[{i}]['value'] must be between 0.0 and 1.0, got {val}")

    # ------------------------------------------------------------------
    def add_value(self, norm_time: float, value: float) -> None:
        if norm_time < 0 or norm_time > 1:
            raise SyntaxError(f"invalid normTime, must be between 0 and 1: {norm_time}")
        if value < 0 or value > 1:
            raise SyntaxError(f"invalid value, must be between 0 and 1: {value}")

        idx = next((i for i, v in enumerate(self.values) if v['norm_time'] == norm_time), -1)
        if idx != -1:
            self.values[idx]['value'] = value
        else:
            self.values.append({'norm_time': norm_time, 'value': value})
            self.values.sort(key=lambda x: x['norm_time'])

    # ------------------------------------------------------------------
    def remove_value(self, idx: int) -> None:
        if idx < 0 or idx > len(self.values) - 1:
            raise SyntaxError(f"invalid idx, must be between 0 and {len(self.values) - 1}: {idx}")
        if idx == 0 or idx == len(self.values) - 1:
            raise SyntaxError("cannot remove first or last value")
        self.values.pop(idx)

    # ------------------------------------------------------------------
    def value_at_x(self, x: float) -> float:
        if x < 0 or x > 1:
            raise SyntaxError(f"invalid x, must be between 0 and 1: {x}")
        idx = -1
        for i in range(len(self.values) - 1, -1, -1):
            if self.values[i]['norm_time'] <= x:
                idx = i
                break
        if idx == -1:
            raise SyntaxError(f"invalid x, must be between 0 and 1: {x}")
        elif idx == len(self.values) - 1:
            return self.values[idx]['value']
        else:
            start = self.values[idx]
            end = self.values[idx + 1]
            slope = (end['value'] - start['value']) / (end['norm_time'] - start['norm_time'])
            return start['value'] + slope * (x - start['norm_time'])

    # ------------------------------------------------------------------
    def generate_value_curve(self, value_dur: float, duration: float, max_val: float = 1.0) -> List[float]:
        value_ct = round(duration / value_dur)
        self.values.sort(key=lambda x: x['norm_time'])
        norm_times = [i / value_ct for i in range(value_ct + 1)]
        return [max_val * self.value_at_x(nt) for nt in norm_times]

    # ------------------------------------------------------------------
    def partition(self, dur_array: List[float]) -> List['Automation']:
        starts = get_starts(dur_array)
        ends = get_ends(dur_array)
        new_automations: List[Automation] = []
        for start, end in zip(starts, ends):
            start_val = self.value_at_x(start)
            end_val = self.value_at_x(end)
            new_automations.append(
                Automation({'values': [
                    {'norm_time': 0.0, 'value': start_val},
                    {'norm_time': 1.0, 'value': end_val}
                ]})
            )
        for v in self.values:
            if v['norm_time'] not in starts and v['norm_time'] not in ends:
                for i in range(len(starts)):
                    if starts[i] < v['norm_time'] < ends[i]:
                        dur = ends[i] - starts[i]
                        rel_norm_time = (v['norm_time'] - starts[i]) / dur if dur != 0 else 0
                        new_automations[i].add_value(rel_norm_time, v['value'])
        return new_automations

    # ------------------------------------------------------------------
    @staticmethod
    def compress(automations: List['Automation'], dur_array: List[float]) -> 'Automation':
        all_values: List[AutomationValueType] = []
        dur_accumulator = 0.0
        for a, dur in zip(automations, dur_array):
            rel_values = [
                {'norm_time': v['norm_time'] * dur + dur_accumulator, 'value': v['value']}
                for v in a.values
            ]
            all_values.extend(rel_values)
            dur_accumulator += dur
        unique: List[AutomationValueType] = []
        seen = set()
        for v in all_values:
            if v['norm_time'] not in seen:
                unique.append(v)
                seen.add(v['norm_time'])
        all_values = unique
        changed = True
        while changed:
            changed = False
            for i in range(len(all_values) - 2):
                a = all_values[i]
                b = all_values[i + 1]
                c = all_values[i + 2]
                slope1 = (b['value'] - a['value']) / (b['norm_time'] - a['norm_time'])
                slope2 = (c['value'] - b['value']) / (c['norm_time'] - b['norm_time'])
                if close_to(slope1, slope2):
                    all_values.pop(i + 1)
                    changed = True
                    break
        return Automation({'values': all_values})

    # ------------------------------------------------------------------
    @staticmethod
    def from_json(obj: Dict) -> 'Automation':
        return Automation(obj)

    def to_json(self) -> Dict:
        return {'values': self.values}
